from typing import Any, Dict, List

import numpy as np
import pytest
import torch as th
import torch.nn as nn

from ....data.containers.tabular_container import TabularContainer
from ....data.extractors.tabular_extractor import TabularFeatureExtractor
from ....data.handler.data_processor import DataProcessor
from ....data.preprocessors.base_preprocessor import StaticPreprocessor
from ....data.tests.iterators.dummy_udao_iterator import DummyUdaoIterator
from ....model.utils.utils import set_deterministic_torch
from ....utils.interfaces import UdaoInput
from ....utils.logging import logger
from ... import concepts as co
from ...soo.mogd import MOGD


class SimpleModel1(nn.Module):
    def forward(self, x: UdaoInput) -> th.Tensor:
        return x.features[:, :1]


class SimpleModel2(nn.Module):
    def forward(self, x: UdaoInput) -> th.Tensor:
        return x.features[:, 1:]


@pytest.fixture()
def data_processor() -> DataProcessor:
    class TabularFeaturePreprocessor(StaticPreprocessor):
        def preprocess(self, tabular_feature: TabularContainer) -> TabularContainer:
            tabular_feature.data.loc[:, "v1"] = tabular_feature.data["v1"] / 1
            tabular_feature.data.loc[:, "v2"] = tabular_feature.data["v2"] - 2
            return tabular_feature

        def inverse_transform(
            self, tabular_feature: TabularContainer
        ) -> TabularContainer:
            tabular_feature.data.loc[:, "v1"] = tabular_feature.data["v1"] * 1
            tabular_feature.data.loc[:, "v2"] = tabular_feature.data["v2"] + 2
            return tabular_feature

    return DataProcessor(
        iterator_cls=DummyUdaoIterator,
        feature_extractors={
            "tabular_features": TabularFeatureExtractor(
                columns=["v1", "v2"],
            ),
            "objectives": TabularFeatureExtractor(columns=["objective_input"]),
        },
        feature_preprocessors={"tabular_features": [TabularFeaturePreprocessor()]},
    )


@pytest.fixture()
def mogd() -> MOGD:
    set_deterministic_torch(42)

    params = MOGD.Params(
        learning_rate=0.1,
        max_iters=100,
        patience=30,
        multistart=2,
        objective_stress=10,
    )
    mogd = MOGD(params)

    return mogd


@pytest.fixture()
def data_processor_paper() -> DataProcessor:
    return DataProcessor(
        iterator_cls=DummyUdaoIterator,
        feature_extractors={
            "tabular_features": TabularFeatureExtractor(
                columns=["v1"],
            ),
            "objectives": TabularFeatureExtractor(columns=["objective_input"]),
        },
    )


class PaperModel1(nn.Module):
    def forward(
        self, input_variables: Dict[str, th.Tensor], input_parameters: Any = None
    ) -> th.Tensor:
        return th.reshape(2400 / (input_variables["cores"]), (-1, 1))


class PaperModel2(nn.Module):
    def forward(
        self, input_variables: Dict[str, th.Tensor], input_parameters: Any = None
    ) -> th.Tensor:
        return th.reshape(input_variables["cores"], (-1, 1))


@pytest.fixture()
def paper_mogd() -> MOGD:
    set_deterministic_torch(42)

    params = MOGD.Params(
        learning_rate=0.1,
        max_iters=100,
        patience=30,
        multistart=2,
        objective_stress=10,
    )
    mogd = MOGD(params)

    return mogd


class TestMOGD:
    @pytest.mark.parametrize(
        "gpu, strict_rounding, expected_obj, expected_vars",
        [
            (False, True, 1, {"v1": 1.0, "v2": 3.0}),
            (False, False, 1, {"v1": 1.0, "v2": 3.0}),
            (True, True, 1, {"v1": 1.0, "v2": 3.0}),
            (True, False, 1, {"v1": 1.0, "v2": 3.0}),
        ],
    )
    def test_solve(
        self,
        mogd: MOGD,
        data_processor: DataProcessor,
        gpu: bool,
        strict_rounding: bool,
        expected_obj: float,
        expected_vars: Dict[str, float],
    ) -> None:
        mogd.device = th.device("cuda") if gpu else th.device("cpu")

        if gpu and not th.cuda.is_available():
            pytest.skip("Skip GPU test")
        set_deterministic_torch(0)
        mogd.strict_rounding = strict_rounding

        problem = co.SOProblem(
            data_processor=data_processor,
            objective=co.Objective(
                "obj1",
                minimize=False,
                function=SimpleModel1(),
                lower=0,
                upper=2,
            ),
            variables={"v1": co.FloatVariable(0, 1), "v2": co.IntegerVariable(2, 3)},
            constraints=[
                co.Constraint(
                    lower=0,
                    upper=1,
                    function=SimpleModel2(),
                )
            ],
        )
        optimal_obj, optimal_vars = mogd.solve(problem, seed=0)
        assert optimal_obj is not None
        np.testing.assert_array_almost_equal(optimal_obj, expected_obj, decimal=5)
        assert optimal_vars == expected_vars

    @pytest.mark.parametrize(
        "strict_rounding, variable, batch_size, expected_obj, expected_vars",
        [
            (True, co.IntegerVariable(0, 24), 1, 200, {"cores": 12}),
            # the strict rounding pulls the gradient descent back to integer value
            # at the end of each iteration. So the variable might not be updated
            # properly if the learning rate is not big enough.
            # This can be improved by data normalization or disabling strict_rounding.
            (True, co.IntegerVariable(0, 24), 16, 150, {"cores": 16}),
            (True, co.FloatVariable(0, 24), 1, 148.17662, {"cores": 16.2}),
            (True, co.FloatVariable(0, 24), 16, 148.17662, {"cores": 16.2}),
            (False, co.IntegerVariable(0, 24), 1, 150, {"cores": 16}),
            (False, co.IntegerVariable(0, 24), 16, 150, {"cores": 16}),
            (False, co.FloatVariable(0, 24), 1, 148.17662, {"cores": 16.2}),
            (False, co.FloatVariable(0, 24), 16, 148.17662, {"cores": 16.2}),
        ],
    )
    def test_solve_paper(
        self,
        paper_mogd: MOGD,
        strict_rounding: bool,
        variable: co.Variable,
        batch_size: int,
        expected_obj: float,
        expected_vars: Dict[str, float],
    ) -> None:
        problem = co.SOProblem(
            objective=co.Objective(
                "obj1",
                minimize=True,
                function=PaperModel1(),
                lower=100,
                upper=200,
            ),
            variables={"cores": variable},
            constraints=[
                co.Objective(
                    name="obj2",
                    lower=8,
                    upper=16.2,
                    function=PaperModel2(),
                    minimize=False,
                )
            ],
        )
        paper_mogd.batch_size = batch_size
        paper_mogd.strict_rounding = strict_rounding
        optimal_obj, optimal_vars = paper_mogd.solve(problem, seed=0)
        logger.debug(f"optimal_obj: {optimal_obj}, optimal_vars: {optimal_vars}")
        assert optimal_obj is not None
        np.testing.assert_allclose([optimal_obj], [expected_obj], rtol=1e-3)
        assert optimal_vars is not None
        assert len(optimal_vars) == 1
        np.testing.assert_allclose(
            [optimal_vars["cores"]], [expected_vars["cores"]], rtol=1e-3
        )

    def test_solve_no_constraints(
        self, mogd: MOGD, data_processor: DataProcessor
    ) -> None:
        def objective_function(x: UdaoInput) -> th.Tensor:
            return th.reshape(x.features[:, 0] ** 2 + x.features[:, 1] ** 2, (-1, 1))

        problem = co.SOProblem(
            data_processor=data_processor,
            objective=co.Objective(
                name="obj1",
                minimize=False,
                function=objective_function,
            ),
            variables={"v1": co.FloatVariable(0, 1), "v2": co.IntegerVariable(2, 3)},
            constraints=[],
        )
        optimal_obj, optimal_vars = mogd.solve(problem, seed=0)

        assert optimal_obj is not None
        np.testing.assert_array_equal(optimal_obj, 2)
        assert optimal_vars is not None
        assert optimal_vars == {"v1": 1, "v2": 3}

    def test_get_unprocessed_input_values(
        self, mogd: MOGD, data_processor: DataProcessor
    ) -> None:
        mogd.batch_size = 4

        input_values, _ = mogd._get_unprocessed_input_values(
            numeric_variables={
                "v1": co.FloatVariable(0, 1),
                "v2": co.IntegerVariable(2, 3),
            },
        )
        assert set(input_values.keys()) == {"v1", "v2"}
        assert all(input_values["v1"] <= 1) and all(input_values["v1"] >= 0)
        assert all([v in [2, 3] for v in input_values["v2"]])

    def test_get_processed_input_values(
        self, mogd: MOGD, data_processor: DataProcessor
    ) -> None:
        mogd.batch_size = 4
        (
            input_values,
            input_shape,
            make_tabular_container,
        ) = mogd._get_processed_input_values(
            data_processor=data_processor,
            numeric_variables={
                "v1": co.FloatVariable(0, 1),
                "v2": co.IntegerVariable(2, 3),
            },
        )
        assert input_values.features.shape == (4, 2)
        assert th.all(input_values.features <= 1) and th.all(input_values.features >= 0)
        assert input_shape.output_names == ["objective_input"]
        assert input_shape.feature_names == ["v1", "v2"]
        container = make_tabular_container(input_values.features)
        np.testing.assert_equal(
            container.data["v1"].values, input_values.features[:, 0].cpu().numpy()
        )
        np.testing.assert_equal(
            container.data["v2"].values, input_values.features[:, 1].cpu().numpy()
        )

    def test_get_processed_input_bounds(
        self, mogd: MOGD, data_processor: DataProcessor
    ) -> None:
        input_lower, input_upper = mogd._get_processed_input_bounds(
            data_processor=data_processor,
            numeric_variables={
                "v1": co.FloatVariable(0, 1),
                "v2": co.IntegerVariable(2, 3),
            },
        )
        assert th.equal(input_lower.features[0].cpu(), th.tensor([0, 0]).cpu())
        assert th.equal(input_upper.features[0].cpu(), th.tensor([1, 1]).cpu())

    def test_get_unprocessed_input_bounds(
        self, mogd: MOGD, data_processor: DataProcessor
    ) -> None:
        data_processor.feature_processors = {}
        input_lower, input_upper = mogd._get_unprocessed_input_bounds(
            numeric_variables={
                "v1": co.FloatVariable(0, 1),
                "v2": co.IntegerVariable(2, 3),
            },
        )
        assert input_lower == {"v1": 0, "v2": 2}
        assert input_upper == {"v1": 1, "v2": 3}

    @pytest.mark.parametrize(
        "objective_values, expected_loss",
        [
            # 0.5 /2 (normalized) * direction (-1 for max) = -0.25
            (th.tensor([0.5]), th.tensor([-0.25])),
            # (-0.2 / 2 - 0.5)**2 + stress (0.1) = 0.46
            (th.tensor([-0.2]), th.tensor([0.46])),
            (th.tensor([[0.5], [0.3], [-0.2]]), th.tensor([[-0.25], [-0.15], [0.46]])),
        ],
    )
    def test__objective_loss_with_bounds(
        self,
        mogd: MOGD,
        data_processor: DataProcessor,
        objective_values: th.Tensor,
        expected_loss: th.Tensor,
    ) -> None:
        objective = co.Objective(
            "obj1",
            minimize=False,
            function=SimpleModel1(),
            lower=0,
            upper=2,
        )
        mogd.objective_stress = 0.1
        loss = mogd.objective_loss(
            objective_values.to(mogd.device), objective.to(mogd.device)  # type: ignore
        )
        # 0.5 /2 (normalized) * direction (-1 for max) = -0.25
        assert th.equal(loss.cpu(), expected_loss.cpu())

    @pytest.mark.parametrize(
        "objective_values, expected_loss",
        [
            # direction * 0.5**2
            (th.tensor([0.5]), th.tensor([-0.25])),
            # direction * (-O.2)**2
            (th.tensor([-0.2]), th.tensor([0.04])),
            (th.tensor([[0.5], [0.3], [-0.2]]), th.tensor([[-0.25], [-0.09], [0.04]])),
        ],
    )
    def test__objective_loss_without_bounds(
        self,
        mogd: MOGD,
        data_processor: DataProcessor,
        objective_values: th.Tensor,
        expected_loss: th.Tensor,
    ) -> None:
        objective = co.Objective(
            "obj1",
            minimize=False,
            function=SimpleModel1(),
        )
        loss = mogd.objective_loss(
            objective_values.to(mogd.device), objective.to(mogd.device)  # type: ignore
        )
        # 0.5 /2 (normalized) * direction (-1 for max) = -0.25
        assert th.allclose(loss.cpu(), expected_loss.cpu())

    @pytest.mark.parametrize(
        "constraint_values, expected_loss",
        [
            (
                [th.tensor([1.1]), th.tensor([1.1]), th.tensor([3.5])],
                # 0.6**2 + 10+ 0 + 0.5**2 + 1000
                th.tensor([1010.61]),
            ),
        ],
    )
    def test_constraints_loss(
        self,
        mogd: MOGD,
        data_processor: DataProcessor,
        constraint_values: List[th.Tensor],
        expected_loss: th.Tensor,
    ) -> None:
        mogd.constraint_stress = 1000
        constraints = [
            co.Objective(
                name="objA",
                minimize=False,
                lower=0,
                upper=1,
                function=SimpleModel1(),
            ),
            co.Constraint(
                lower=0,
                upper=2,
                function=SimpleModel2(),
            ),
            co.Constraint(
                upper=3,
                function=SimpleModel2(),
            ),
        ]
        loss = mogd.constraints_loss(
            [c.to(mogd.device) for c in constraint_values],
            [c.to(mogd.device) for c in constraints],  # type: ignore
        )
        assert th.allclose(loss.cpu(), expected_loss.cpu())

    def test_get_meshed_categorical_variables(self, mogd: MOGD) -> None:
        variables = {
            "v1": co.IntegerVariable(2, 3),
            "v2": co.EnumVariable([4, 5]),
            "v3": co.EnumVariable([10, 20]),
        }
        meshed_variables = mogd.get_meshed_categorical_vars(variables=variables)
        assert meshed_variables is not None
        np.testing.assert_array_equal(
            meshed_variables, [[4.0, 10.0], [5.0, 10.0], [4.0, 20.0], [5.0, 20.0]]
        )



from .base_predicate_embedder import BasePredicateEmbedder
from .doc2vec_embedder import Doc2VecEmbedder, Doc2VecParams
from .word2vec_embedder import Word2VecEmbedder, Word2VecParams

__all__ = [
    "BasePredicateEmbedder",
    "Doc2VecEmbedder",
    "Doc2VecParams",
    "Word2VecEmbedder",
    "Word2VecParams",
]



import re
import warnings
from collections import defaultdict
from dataclasses import dataclass, fields
from typing import Dict, List, Optional, Tuple, cast

import dgl
import networkx as nx
import numpy as np
import torch as th
import torch.nn.functional as F
from networkx.algorithms import isomorphism
from scipy import sparse as sp


def format_size(size: str) -> float:
    unit_to_multiplier = {
        "B": 1,
        "KiB": 1024,
        "MiB": 1024**2,
        "GiB": 1024**3,
        "TiB": 1024**4,
    }
    n, unit = size.replace(",", "").split()

    try:
        return float(n) * unit_to_multiplier[unit]
    except KeyError:
        raise Exception(f"unkown {n} in {size}")


@dataclass
class QueryPlanOperationFeatures:
    """Features of the operations in the logical plan"""

    rows_count: List[float]
    size: List[float]

    @property
    def operation_ids(self) -> List[int]:
        return list(range(len(self.rows_count)))

    @property
    def features_dict(self) -> Dict[str, List]:
        return {"rows_count": self.rows_count, "size": self.size}

    @classmethod
    def get_feature_names_and_types(cls) -> Dict[str, type]:
        """Retrieve the names and element types of features."""
        feature_info = {}
        for f in fields(cls):
            # Assuming the type is always List[SomeType]
            try:
                element_type = f.type.__args__[0]
            except AttributeError:
                raise ValueError(f"Feature {f.name} should be a List, but got {f.type}")
            feature_info[f.name] = element_type
        return feature_info


class QueryPlanStructure:
    """Generate a graph from the tree structure of the logical plan.

    Parameters
    ----------
    node_names : List[str]
        list of the names of the nodes in the graph
    incoming_ids : List[int]
        For each edge i, incoming_ids[i] is the id of the source node
    outgoing_ids : List[int]
        for each edge i, outgoing_ids[i] is the id of the destination node
    """

    def __init__(
        self,
        node_names: List[str],
        incoming_ids: List[int],
        outgoing_ids: List[int],
    ) -> None:
        self.incoming_ids = incoming_ids
        self.outgoing_ids = outgoing_ids
        self.node_id2name = {i: n for i, n in enumerate(node_names)}
        self.graph: dgl.DGLGraph = dgl.graph(
            (incoming_ids, outgoing_ids), num_nodes=len(node_names)
        )
        self._nx_graph: Optional[nx.Graph] = None

    @property
    def nx_graph(self) -> nx.Graph:
        if self._nx_graph is None:
            self._nx_graph = dgl.to_networkx(self.graph)
            nx.set_node_attributes(self._nx_graph, self.node_id2name, name="nname")
        return self._nx_graph

    def graph_match(self, plan: "QueryPlanStructure") -> bool:
        """Checks if the current plan and another one are isomorphic.

        Parameters
        ----------
        plan : QueryPlanStructure
            another plan to compare with

        Returns
        -------
        bool
            True if the plans are isomorphic, False otherwise
        """
        matcher = isomorphism.GraphMatcher(
            self.nx_graph,
            plan.nx_graph,
            node_match=lambda n1, n2: n1["nname"] == n2["nname"],
        )
        return matcher.is_isomorphic()  # type: ignore


class _LogicalOperation:
    """Describes an operation in the logical plan, with associated features.

    Parameters
    ----------
    id : int
        id of the operation in the query plan (one id per line in the plan)
    value : str
        text of the operation (line in the string representation of the plan)
    """

    def __init__(self, id: int, value: str) -> None:
        self.id: int = id
        self.value: str = value
        self._rank: Optional[float] = None

    id: int
    value: str
    _rank: Optional[float] = None

    @property
    def rank(self) -> float:
        """Compute the rank of the operation,
        i.e. the number of spaces before the operation name."""
        if self._rank is None:
            self._rank = (len(self.value) - len(self.value.lstrip())) / 3
        return self._rank

    @rank.setter
    def rank(self, rank: float) -> None:
        self._rank = rank

    @property
    def link(self) -> str:
        """Compute the link between the current operation and the previous one,
        i.e. the character before the operation name."""
        return self.value.lstrip()[0]

    @property
    def name(self) -> str:
        """Compute the name of the operation: the first word after the link,
        or with additional term after the space in the case of a Relation operation"""
        _, b, c = re.split(r"([a-z])", self.value, 1, flags=re.I)
        name = (b + c).split(" ")[0]
        if name == "Relation":
            name = (b + c).split("[")[0]
        return name

    @property
    def size(self) -> float:
        """Extract the size of the operation, in bytes"""
        return format_size(self.value.split("sizeInBytes=")[1].split("B")[0] + "B")

    def get_rows_count(
        self, id_predecessors: List[int], rows_count: List[float]
    ) -> float:
        """Extract the number of rows of the operation,
        or infer it from its predecessors if not available"""
        if "rowCount=" in self.value:
            return float(self.value.split("rowCount=")[1].split(")")[0])
        else:
            if len(id_predecessors) == 1:
                pid = id_predecessors[0]
                if rows_count[pid] is None:
                    raise ValueError(
                        "The number of rows of the parent should be a float."
                        f"Got None for {pid}"
                    )
                return rows_count[pid]
            elif len(id_predecessors) == 2:
                pid1, pid2 = id_predecessors
                if rows_count[pid1] is None or rows_count[pid2] is None:
                    raise ValueError(
                        "The number of rows of the parent should be a float, but "
                        f"got None for {pid1}: {rows_count[pid1]}"
                        f"or {pid2}: {rows_count[pid2]}"
                    )
                if rows_count[pid1] * rows_count[pid2] != 0:
                    raise ValueError(
                        "The product of number of rows of parents should be 0, but "
                        f"got {rows_count[pid1] * rows_count[pid2]}"
                    )
                return 0
            else:
                raise NotImplementedError("More than 2 predecessors")

    def get_unindented(self) -> "_LogicalOperation":
        """Return the operation without one level of indentation removed"""
        return _LogicalOperation(self.id, self.value.lstrip()[3:])


def _get_tree_structure(
    operations: List[_LogicalOperation],
) -> Tuple[List[int], List[int]]:
    """Define the tree structure of the logical plan, as a tuple of two lists,
    where U[i], V[i] are the incoming and outgoing nodes of edge i.

    Parameters
    ----------
    steps : List[LogicalStep]
        List of operation in the logical plan

    Returns
    -------
    Tuple[List[int], List[int]]
        _description_

    Raises
    ------
    ValueError
        _description_
    """
    incoming_ids: List[int] = []
    outgoing_ids: List[int] = []
    pre_id = operations[0].id
    operations[0].rank = -1
    threads: Dict[float, List[_LogicalOperation]] = defaultdict(list)
    for i, op in enumerate(operations):
        if i == 0:
            continue
        if op.link == ":":
            unindented_step = op.get_unindented()
            threads[op.rank].append(unindented_step)

        elif op.link == "+":
            incoming_ids.append(op.id)
            outgoing_ids.append(pre_id)
            pre_id = op.id
        else:
            raise ValueError(
                f"Unexpected link {op.link} in {op.value}. Expected : or +"
            )

    for rank, rank_operations in threads.items():
        incoming_ids.append(rank_operations[0].id)
        outgoing_id_candidates = [
            op.id for op in operations if op.rank == rank - 1 and op.link != ":"
        ]
        if len(outgoing_id_candidates) != 1:
            warnings.warn(
                f"More than one candidate outgoing id for rank {rank}."
                "\n Taking the first candidate."
            )
        outgoing_ids.append(outgoing_id_candidates[0])
        sub_from_ids, sub_to_ids = _get_tree_structure(rank_operations)
        incoming_ids += sub_from_ids
        outgoing_ids += sub_to_ids
    return incoming_ids, outgoing_ids


def compute_meta_features(
    structure: QueryPlanStructure, features: QueryPlanOperationFeatures
) -> Dict[str, float]:
    """Compute the meta features of the operations in the query plan."""
    graph = structure.graph
    input_meta = {}
    for op_feature, feature_values in features.features_dict.items():
        graph.ndata[op_feature] = th.tensor(feature_values)
    input_nodes_index = th.where(graph.in_degrees() == 0)[0]
    for op_feature in features.features_dict.keys():
        input_meta[f"meta_{op_feature}"] = (
            graph.ndata[op_feature][input_nodes_index].sum(0).item()
        )
        del graph.ndata[op_feature]
    return input_meta


def get_laplacian_positional_encoding(
    graph: dgl.DGLGraph, pos_enc_dim: Optional[int] = None
) -> th.Tensor:
    """
    Graph positional encoding v/ Laplacian eigenvectors

    Parameters
    ----------
    graph : dgl.DGLGraph
        graph to encode
    pos_enc_dim : Optional[int], optional
        number of eigenvectors to use,
        by default set to the number of nodes in the graph - 2
    """
    if pos_enc_dim is None:
        pos_enc_dim = graph.number_of_nodes() - 2
    # Laplacian
    # A = g.adjacency_matrix_scipy(return_edge_ids=False).astype(float)
    A = graph.adj_external(scipy_fmt="csr").T  # adjacency matrix

    in_degrees = cast(th.Tensor, graph.in_degrees()).numpy().clip(1)
    N = sp.diags(in_degrees**-0.5, dtype=float, format="csr")
    L = sp.eye(graph.number_of_nodes()) - N * A * N

    # Eigenvectors with numpy
    eigen_values, eigen_vectors = np.linalg.eig(L.toarray())
    idx = eigen_values.argsort()  # increasing order
    eigen_values, eigen_vectors = eigen_values[idx], np.real(eigen_vectors[:, idx])
    return th.from_numpy(eigen_vectors[:, 1 : pos_enc_dim + 1]).float()


def add_positional_encoding(graph: dgl.DGLGraph, size: int) -> dgl.DGLGraph:
    """Add positional encoding to the graph.

    Parameters
    ----------
    graph : dgl.DGLGraph
        graph to encode
    size : int
        size of the positional encoding
        (truncated if the graph has less nodes or padded with 0 otherwise)

    Returns
    -------
    dgl.DGLGraph
        graph with positional encoding as "pos_enc" node feature
    """
    bidirectional = cast(dgl.DGLGraph, dgl.to_bidirected(graph))
    graph.ndata["pos_enc"] = resize_tensor(
        get_laplacian_positional_encoding(bidirectional, graph.num_nodes() - 2), size
    )
    return graph


def resize_tensor(tensor: th.Tensor, new_size: int) -> th.Tensor:
    """Pad or truncate the last dimension of a tensor to a given size."""
    current_size = tensor.shape[-1]
    if current_size < new_size:
        tensor = F.pad(
            tensor,
            pad=(0, new_size - current_size),
            mode="constant",
            value=0,
        )
    else:
        # truncate last dimension
        tensor = tensor[..., :new_size]

    return tensor


def random_flip_positional_encoding(graph: dgl.DGLGraph) -> dgl.DGLGraph:
    """Add same random sign flip to all nodes positional encodings of the graph."""
    pos_enc = cast(th.Tensor, graph.ndata["pos_enc"])
    sign_flip = th.rand(pos_enc.size(1))
    sign_flip[sign_flip >= 0.5] = 1.0
    sign_flip[sign_flip < 0.5] = -1.0
    pos_enc *= sign_flip.unsqueeze(0)
    graph.ndata["pos_enc"] = pos_enc
    return graph


def extract_query_plan_features(
    logical_plan: str,
) -> Tuple[QueryPlanStructure, QueryPlanOperationFeatures, Dict[str, float]]:
    """Extract:
    - features of the operations in the logical plan
    - meta features aggregated from operations of the logical plan
    - the tree structure of the logical plan


    Parameters
    ----------
    logical_plan : str
        string representation of the logical plan

    Returns
    -------
    Tuple[QueryPlanStructure, QueryPlanOperationFeatures, Dict[str, float]]
        Query plan structure (dgl graph with node names)
        Query plan features (sizes and rows_counts of the operations)
        Meta features (aggregated features of the operations)
    """
    operations = [
        _LogicalOperation(id=i, value=step)
        for i, step in enumerate(logical_plan.splitlines())
    ]
    node_names: List[str] = []
    sizes: List[float] = []
    rows_counts: List[float] = []

    incoming_ids, outgoing_ids = _get_tree_structure(operations)
    predecessors: Dict[int, List[int]] = defaultdict(list)
    for f, t in zip(incoming_ids, outgoing_ids):
        predecessors[t].append(f)

    for step in reversed(operations):
        node_names.insert(0, step.name)
        sizes.insert(0, step.size)
        rows_counts.insert(0, step.get_rows_count(predecessors[step.id], rows_counts))

    op_features = QueryPlanOperationFeatures(rows_count=rows_counts, size=sizes)
    structure = QueryPlanStructure(
        node_names=node_names, incoming_ids=incoming_ids, outgoing_ids=outgoing_ids
    )
    meta_features = compute_meta_features(structure, op_features)
    return structure, op_features, meta_features



import pandas as pd
from sklearn.preprocessing import MinMaxScaler

from ...containers.tabular_container import TabularContainer
from ...preprocessors.normalize_preprocessor import NormalizePreprocessor


class TestNormalizePreprocessor:
    def test_fit_and_transform(self) -> None:
        df_train = pd.DataFrame.from_dict({"index": [1, 2], "col": [3, 4]})
        df_val = pd.DataFrame.from_dict({"index": [3, 4], "col": [5, 6]})
        df_train.set_index("index", inplace=True)
        df_val.set_index("index", inplace=True)
        train_container = TabularContainer(df_train)
        processor: NormalizePreprocessor[TabularContainer] = NormalizePreprocessor(
            MinMaxScaler()
        )
        train_container_result = processor.preprocess(train_container, "train")
        assert train_container_result.data.values.tolist() == [[0.0], [1.0]]
        assert train_container.data.values.tolist() == [[3.0], [4.0]]
        val_container = TabularContainer(df_val)
        val_container_result = processor.preprocess(val_container, "val")
        assert val_container_result.data.values.tolist() == [[2.0], [3.0]]
        assert val_container.data.values.tolist() == [[5.0], [6.0]]

    def test_inverse_transform(self) -> None:
        df_train = pd.DataFrame.from_dict({"index": [1, 2], "col": [3, 4]})
        df_val = pd.DataFrame.from_dict({"index": [3, 4], "col": [5, 6]})
        df_train.set_index("index", inplace=True)
        df_val.set_index("index", inplace=True)
        train_container = TabularContainer(df_train)
        processor: NormalizePreprocessor[TabularContainer] = NormalizePreprocessor(
            MinMaxScaler()
        )
        train_container_result = processor.preprocess(train_container, "train")
        assert train_container_result.data.values.tolist() == [[0.0], [1.0]]
        val_container = TabularContainer(df_val)
        val_container_result = processor.preprocess(val_container, "val")
        assert val_container_result.data.values.tolist() == [[2.0], [3.0]]
        val_container_result_inverse = processor.inverse_transform(val_container_result)
        assert val_container_result_inverse.data.values.tolist() == [[5.0], [6.0]]
        assert val_container.data.values.tolist() == [[5.0], [6.0]]

    def test_inverse_transform_only_applies_to_numeric(self) -> None:
        df_train = pd.DataFrame.from_dict(
            {"index": [1, 2], "col": [3, 4], "col2": [1, 2]}
        )
        df_val = pd.DataFrame.from_dict(
            {"index": [3, 4], "col": [5, 6], "col2": [3, 4]}
        )
        df_train["col2"] = df_train["col2"].astype("category")
        df_val["col2"] = df_val["col2"].astype("category")
        df_train.set_index("index", inplace=True)
        df_val.set_index("index", inplace=True)
        train_container = TabularContainer(df_train)
        processor: NormalizePreprocessor[TabularContainer] = NormalizePreprocessor(
            MinMaxScaler()
        )
        # Test that inverse transform applies correctly to numeric columns
        train_container_result = processor.preprocess(train_container, "train")
        train_container_result_inverse = processor.inverse_transform(
            train_container_result
        )
        assert train_container_result.data["col"].values.tolist() == [0.0, 1.0]
        assert train_container_result_inverse.data["col"].values.tolist() == [3.0, 4.0]
        val_container = TabularContainer(df_val)
        val_container_result = processor.preprocess(val_container, "val")
        assert val_container_result.data["col"].values.tolist() == [2.0, 3.0]
        val_container_result_inverse = processor.inverse_transform(val_container_result)
        assert val_container_result_inverse.data["col"].values.tolist() == [5.0, 6.0]

        # Test that categorical column remains unchanged
        pd.testing.assert_series_equal(
            val_container_result_inverse.data["col2"], val_container_result.data["col2"]
        )
        pd.testing.assert_series_equal(
            val_container_result_inverse.data["col2"], val_container.data["col2"]
        )



from .constraint import Constraint
from .objective import Objective
from .problem import MOProblem, SOProblem
from .utils import InputParameters, InputVariables
from .variable import (
    BoolVariable,
    EnumVariable,
    FloatVariable,
    IntegerVariable,
    NumericVariable,
    Variable,
)

__all__ = [
    "Constraint",
    "Objective",
    "Variable",
    "NumericVariable",
    "EnumVariable",
    "IntegerVariable",
    "BoolVariable",
    "FloatVariable",
    "InputVariables",
    "InputParameters",
    "MOProblem",
    "SOProblem",
]



import numpy as np
import pytest

from ...predicate_embedders import Doc2VecEmbedder, Doc2VecParams


@pytest.fixture
def doc2vec_embedder() -> Doc2VecEmbedder:
    return Doc2VecEmbedder(Doc2VecParams())


class TestDoc2Vec:
    def test_init(self, doc2vec_embedder: Doc2VecEmbedder) -> None:
        assert doc2vec_embedder.d2v_model is not None
        assert doc2vec_embedder._is_trained is False

    def test_fit_sets_is_trained(self, doc2vec_embedder: Doc2VecEmbedder) -> None:
        training_plans = ["a b c", "a b d"]
        doc2vec_embedder.fit(training_plans)
        assert doc2vec_embedder._is_trained is True

    def test_transform_not_trained_raises_error(
        self, doc2vec_embedder: Doc2VecEmbedder
    ) -> None:
        with pytest.raises(ValueError):
            doc2vec_embedder.transform(["a b c"])

    def test_transform_trained_output_values(
        self, doc2vec_embedder: Doc2VecEmbedder
    ) -> None:
        training_plans = ["a b c", "a b d"]
        doc2vec_embedder.fit(training_plans)
        training_encodings = doc2vec_embedder.transform(training_plans)
        encoding = doc2vec_embedder.transform(["a b c", "a b x"])
        dot = np.dot(training_encodings[0], encoding[0])
        norm_a = np.linalg.norm(training_encodings[0])
        norm_b = np.linalg.norm(encoding[0])
        cosine_similarity = dot / (norm_a * norm_b)
        # similary superior to 0.999
        assert cosine_similarity > 0.999



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union

import pandas as pd

from ...data.containers import BaseContainer
from ...data.utils.utils import DatasetType

T = TypeVar("T", bound=BaseContainer)


class TrainedExtractor(ABC, Generic[T]):
    trained: bool = True

    def __init__(self) -> None:
        pass

    @abstractmethod
    def extract_features(self, df: pd.DataFrame, split: DatasetType) -> T:
        pass


class StaticExtractor(ABC, Generic[T]):
    trained: bool = False

    def __init__(self) -> None:
        pass

    @abstractmethod
    def extract_features(self, df: pd.DataFrame) -> T:
        pass


FeatureExtractor = Union[TrainedExtractor, StaticExtractor]



from dataclasses import dataclass
from typing import List, Optional, Sequence

import numpy as np
from gensim.models import Doc2Vec
from gensim.models.doc2vec import TaggedDocument

from .base_predicate_embedder import BasePredicateEmbedder
from .word2vec_embedder import Word2VecParams


@dataclass
class Doc2VecParams(Word2VecParams):
    pass


class Doc2VecEmbedder(BasePredicateEmbedder):
    """A class to embed query plans using Doc2Vec.
    To use it:
    - first call fit_transform on a list of training query plans,
    - then call transform on a list of query plans.

    N.B. To ensure reproducibility, several things need to be done:
    - set the seed in the Doc2VecParams
    - set the PYTHONHASHSEED
    - set the number of workers to 1

    Parameters
    ----------
    d2v_params : Doc2VecParams
        Parameters to pass to the gensim.Doc2Vec model
        (ref: https://radimrehurek.com/gensim/models/doc2vec.html)
    """

    def __init__(self, d2v_params: Optional[Doc2VecParams] = None) -> None:
        if d2v_params is None:
            d2v_params = Doc2VecParams()
        self.d2v_params = d2v_params
        self.d2v_model = Doc2Vec(
            min_count=d2v_params.min_count,
            window=d2v_params.window,
            vector_size=d2v_params.vec_size,
            sample=d2v_params.sample,
            alpha=d2v_params.alpha,
            min_alpha=d2v_params.min_alpha,
            workers=d2v_params.workers,
            seed=d2v_params.seed,
            epochs=d2v_params.epochs,
        )
        self._is_trained = False

    def _prepare_corpus(self, texts: Sequence[str], /) -> List[TaggedDocument]:
        """Transform strings into a list of TaggedDocument

        a TaggedDocument consists in a list of tokens and a tag
        (here the index of the plan)
        """
        tokens_list = list(map(lambda x: x.split(), texts))
        corpus = [TaggedDocument(d, [i]) for i, d in enumerate(tokens_list)]
        return corpus

    def fit(
        self, training_texts: Sequence[str], /, epochs: Optional[int] = None
    ) -> None:
        """Train the Doc2Vec model on the training plans

        Parameters
        ----------
        training_plans : Sequence[str]
            list of training plans
        epochs : int, optional
            number of epochs for training the model, by default
            will use the value in the Doc2VecParams.

        Returns
        -------
        np.ndarray
            Normalized (L2) embeddings of the training plans
        """
        if epochs is None:
            epochs = self.d2v_params.epochs
        corpus = self._prepare_corpus(training_texts)
        self.d2v_model.build_vocab(corpus)
        self.d2v_model.train(
            corpus, total_examples=self.d2v_model.corpus_count, epochs=epochs
        )
        self._is_trained = True

    def transform(self, texts: Sequence[str]) -> np.ndarray:
        """Transform a list of query plans into normalized embeddings.

        Parameters
        ----------
        plans : Sequence[str]
            list of query plans
        epochs: int, optional
            number of epochs for infering a document's embedding,
            by default will use the value in the Doc2VecParams.

        Returns
        -------
        np.ndarray
            Normalized (L2) embeddings of the query plans

        Raises
        ------
        ValueError
            If the model has not been trained
        """
        epochs = self.d2v_params.epochs
        if not self._is_trained:
            raise ValueError("Must call fit_transform before calling transform")
        encodings = [
            self.d2v_model.infer_vector(doc.split(), epochs=epochs) for doc in texts
        ]
        norms = np.linalg.norm(encodings, axis=1)
        # normalize the embeddings
        return encodings / norms[..., np.newaxis]

    def fit_transform(
        self, training_texts: Sequence[str], /, epochs: Optional[int] = None
    ) -> np.ndarray:
        self.fit(training_texts, epochs)
        return self.transform(training_texts)



import itertools
from dataclasses import dataclass, field
from typing import Dict, List, Mapping, Optional

import numpy as np
import torch as th

from ..concepts import EnumVariable, IntegerVariable, NumericVariable, Variable
from ..utils.moo_utils import get_default_device
from .sampler_solver import SamplerSolver


class GridSearchSolver(SamplerSolver):
    """Solving a SOO problem by grid search over variables"""

    @dataclass
    class Params:
        n_grids_per_var: List[int]
        """List of grid sizes for each variable"""
        device: Optional[th.device] = field(default_factory=get_default_device)
        """device on which to perform torch operations, by default available device."""

    def __init__(self, params: Params) -> None:
        """
        :param gs_params: dict, the parameters used in grid_search
        """
        super().__init__(params.device)
        self.n_grids_per_var = params.n_grids_per_var

    def _process_variable(self, var: Variable, n_grids: int) -> np.ndarray:
        """Define grid point in fonction of the variable type"""
        if isinstance(var, NumericVariable):
            # make sure the grid point is the same with the type
            # e.g., if int x.min=0, x.max=5, n_grids_per_var=10,
            # ONLY points[0, 1, 2, 3, 4, 5] are feasible
            if isinstance(var, IntegerVariable):
                if n_grids > (var.upper - var.lower + 1):
                    n_grids = int(var.upper - var.lower + 1)

            var_grid = np.linspace(var.lower, var.upper, num=n_grids, endpoint=True)
            if isinstance(var, IntegerVariable):
                return np.round(var_grid).astype(int)
            return var_grid
        elif isinstance(var, EnumVariable):
            return np.array(var.values)
        else:
            raise NotImplementedError(
                f"ERROR: variable type {type(var)} is not supported!"
            )

    def _get_input(
        self, variables: Mapping[str, Variable], seed: Optional[int] = None
    ) -> Dict[str, np.ndarray]:
        """
        Generate grids for each variable

        Parameters
        ----------
        variables: Mapping[str, Variable]
            variables to generate

        Returns
        -------
        Dict[str, np.ndarray]
            Dict with array of values for each variable
        """
        grids_list = []
        variable_names = list(variables.keys())

        for i, var_name in enumerate(variable_names):
            var = variables[var_name]
            var_n_grids = self.n_grids_per_var[i]
            grids_list.append({var_name: self._process_variable(var, var_n_grids)})

        values_list = [list(d.values())[0] for d in grids_list]
        cartesian_product = np.array([list(i) for i in itertools.product(*values_list)])
        result_dict = {
            var_name: cartesian_product.T[i]
            for i, var_name in enumerate(variable_names)
        }

        return result_dict



from typing import Callable, Optional, Union

import torch as th

from ...utils.interfaces import VarTypes
from ..concepts.utils import UdaoFunction
from .constraint import Constraint


class Objective(Constraint):
    """

    Parameters
    ----------
    name : str
        Name of the objective.
    minimize : bool
        Direction of the objective: if True, minimize, else maximize.
    type: VarTypes
        Type of the objective, by default VarTypes.FLOAT
    """

    def __init__(
        self,
        name: str,
        minimize: bool,
        function: Union[UdaoFunction, th.nn.Module, Callable[..., th.Tensor]],
        lower: Optional[float] = None,
        upper: Optional[float] = None,
        type: VarTypes = VarTypes.FLOAT,
    ):
        super().__init__(function=function, lower=lower, upper=upper)
        self.name = name
        self.minimize = minimize
        self.type = type

    @property
    def direction(self) -> int:
        """Get gradient direction from optimization type"""
        if self.minimize:
            return 1
        else:
            return -1

    def __repr__(self) -> str:
        return (
            f"Objective(name={self.name}, "
            f"direction={'min' if self.minimize else 'max'}, "
            f"lower={self.lower}, upper={self.upper})"
        )



from typing import Dict, Sequence

import pytest
import torch as th
from torch import nn

from ....data.containers.tabular_container import TabularContainer
from ....data.extractors.tabular_extractor import TabularFeatureExtractor
from ....data.handler.data_processor import DataProcessor
from ....data.preprocessors.base_preprocessor import StaticPreprocessor
from ....data.tests.iterators.dummy_udao_iterator import DummyUdaoIterator
from ....utils.interfaces import UdaoEmbedInput
from ...concepts import (
    BoolVariable,
    Constraint,
    FloatVariable,
    IntegerVariable,
    Objective,
    Variable,
)
from ...concepts.problem import MOProblem
from ...concepts.utils import InputParameters, InputVariables
from ...soo.mogd import MOGD


class ObjModel1(nn.Module):
    def forward(self, x: UdaoEmbedInput) -> th.Tensor:
        return th.reshape(x.features[:, 0] ** 2, (-1, 1))


class ObjModel2(nn.Module):
    def forward(self, x: UdaoEmbedInput) -> th.Tensor:
        return th.reshape(x.features[:, 1] ** 2, (-1, 1))


class ComplexObj1(nn.Module):
    def forward(self, x: UdaoEmbedInput) -> th.Tensor:
        return th.reshape(x.features[:, 0] ** 2 - x.features[:, 1] ** 2, (-1, 1))


class ComplexObj2(nn.Module):
    def forward(self, x: UdaoEmbedInput) -> th.Tensor:
        return th.reshape(x.features[:, 0] ** 2 + x.features[:, 1] ** 2, (-1, 1))


class TabularFeaturePreprocessor(StaticPreprocessor):
    def preprocess(self, tabular_feature: TabularContainer) -> TabularContainer:
        tabular_feature.data.loc[:, "v1"] = tabular_feature.data["v1"] / 1
        tabular_feature.data.loc[:, "v2"] = (tabular_feature.data["v2"] - 1) / 6
        return tabular_feature

    def inverse_transform(self, tabular_feature: TabularContainer) -> TabularContainer:
        tabular_feature.data.loc[:, "v1"] = tabular_feature.data["v1"] * 1
        tabular_feature.data.loc[:, "v2"] = tabular_feature.data["v2"] * 6 + 1
        return tabular_feature


@pytest.fixture()
def data_processor() -> DataProcessor:
    return DataProcessor(
        iterator_cls=DummyUdaoIterator,
        feature_extractors={
            "tabular_features": TabularFeatureExtractor(
                columns=["v1", "v2"],
            ),
            "objectives": TabularFeatureExtractor(columns=["objective_input"]),
        },
        feature_preprocessors={"tabular_features": [TabularFeaturePreprocessor()]},
    )


@pytest.fixture
def mogd() -> MOGD:
    return MOGD(
        MOGD.Params(
            learning_rate=0.1,
            max_iters=100,
            patience=10,
            multistart=2,
            objective_stress=10,
            constraint_stress=1e5,
            device=th.device("cpu"),
        )
    )


@pytest.fixture
def two_obj_problem(data_processor: DataProcessor) -> MOProblem:
    objectives = [
        Objective("obj1", minimize=True, function=ObjModel1()),
        Objective("obj2", minimize=True, function=ObjModel2()),
    ]
    variables: Dict[str, Variable] = {
        "v1": FloatVariable(0, 1),
        "v2": IntegerVariable(1, 7),
    }
    constraints: Sequence[Constraint] = []

    return MOProblem(
        objectives=objectives,
        variables=variables,
        constraints=constraints,
        data_processor=data_processor,
        input_parameters=None,
    )


@pytest.fixture
def three_obj_problem(
    two_obj_problem: MOProblem, data_processor: DataProcessor
) -> MOProblem:
    return MOProblem(
        objectives=[
            Objective("obj1", minimize=False, function=ObjModel1()),
            Objective("obj2", minimize=False, function=ObjModel2()),
            Objective("obj3", minimize=False, function=ComplexObj2()),
        ],
        variables=two_obj_problem.variables,
        constraints=two_obj_problem.constraints,
        input_parameters=two_obj_problem.input_parameters,
        data_processor=data_processor,
    )


@pytest.fixture
def simple_problem() -> MOProblem:
    def obj_func1(
        input_variables: InputVariables, input_parameters: InputParameters = None
    ) -> th.Tensor:
        return input_variables["v1"] + (input_parameters or {}).get("count", 0)

    def obj_func2(
        input_variables: InputVariables, input_parameters: InputParameters = None
    ) -> th.Tensor:
        return (input_variables["v1"] + input_variables["v2"]) / 10 + (
            input_parameters or {}
        ).get("count", 0)

    objectives = [
        Objective("obj1", function=obj_func1, minimize=True),
        Objective("obj2", function=obj_func2, minimize=True),
    ]

    def constraint_func(
        input_variables: InputVariables, input_parameters: InputParameters = None
    ) -> th.Tensor:
        return (
            (input_variables["v1"] + input_variables["v2"])
            - 2
            - (input_parameters or {}).get("count", 0)
        )

    constraints = [Constraint(function=constraint_func, lower=0)]
    return MOProblem(
        objectives=objectives,
        constraints=constraints,
        variables={"v1": BoolVariable(), "v2": IntegerVariable(1, 7)},
        input_parameters={"count": 1},
    )



from typing import List, Optional

import torch as th
import torch.nn as nn
import torch.nn.functional as F


class MLPReadout(nn.Module):
    """MLP inner layers for regression.

    Parameters
    ----------
    input_dim : int
        Dimension of the input tensor
    hidden_dim : int
        Dimension of the hidden layers
    output_dim : int
        Dimension of the output tensor
    n_layers : int, optional
        Number of fully connected layers, by default 2
    dropout : float, optional
        Dropout probability, by default 0
        If no dropout is set, batch normalization is used instead.
    agg_dims : Optional[List[int]], optional
        Dimensions of the aggregation layers, by default None
        None means no aggregation layers.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        n_layers: int = 2,
        dropout: float = 0,
        agg_dims: Optional[List[int]] = None,
    ) -> None:
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers
        self.dropout = dropout
        self.agg_dims = agg_dims
        self.FC_layers = self._load_FC_layers()
        self.BN_layers = self._load_BN_layers()

        self.dropout = dropout

    def _load_FC_layers(self) -> nn.ModuleList:
        """Create the list of fully connected layers."""
        list_FC_layers = [nn.Linear(self.input_dim, self.hidden_dim)]
        for _ in range(self.n_layers - 1):
            list_FC_layers.extend([nn.Linear(self.hidden_dim, self.hidden_dim)])
        if self.agg_dims is None:
            list_FC_layers.append(nn.Linear(self.hidden_dim, self.output_dim))
        else:
            agg_input_dim = self.hidden_dim
            for agg_output_dim in self.agg_dims:
                list_FC_layers.append(nn.Linear(agg_input_dim, agg_output_dim))
                agg_input_dim = agg_output_dim
            list_FC_layers.append(nn.Linear(agg_input_dim, self.output_dim))
        return nn.ModuleList(list_FC_layers)

    def _load_BN_layers(self) -> Optional[nn.ModuleList]:
        """
        Create the list of batch normalization layers,
        if dropout is not used.
        One for each FC layer except last one.
        One for each aggregation layer.
        """
        if self.dropout == 0:
            list_BN_layers = [nn.BatchNorm1d(self.hidden_dim)]
            list_BN_layers.extend(
                [nn.BatchNorm1d(self.hidden_dim) for _ in range(self.n_layers - 1)]
            )
            if self.agg_dims is not None:
                list_BN_layers.extend([nn.BatchNorm1d(a) for a in self.agg_dims])
            return nn.ModuleList(list_BN_layers)
        return None

    def forward(self, input: th.Tensor) -> th.Tensor:
        y = input
        for layer_id in range(len(self.FC_layers) - 1):
            y = self.FC_layers[layer_id](y)
            y = F.relu(y)
            if self.BN_layers is not None:
                y = self.BN_layers[layer_id](y)
            else:
                y = F.dropout(y, self.dropout, training=self.training)
        return self.FC_layers[-1](y)



from typing import Any, Dict, List, Optional

import dgl
import torch as th
import torch.nn as nn
import torch.nn.functional as F

from .multi_head_attention import (
    ATTENTION_TYPES,
    AttentionLayerName,
    QFMultiHeadAttentionLayer,
    RAALMultiHeadAttentionLayer,
)


class GraphTransformerLayer(nn.Module):
    """Graph Transformer Layer that applies multi-head attention
    and feed-forward network to an input containing a graph.
    proposed by "A Generalization of Transformer Networks to Graphs", DLG-AAAI'21.
    https://arxiv.org/pdf/2012.09699.pdf

    Parameters
    ----------
    in_dim : int
        Dimension of the input tensor
    out_dim : int
        Dimension of the output tensor
    n_heads : int
        Number of attention heads
    dropout : float, optional
        Probability of dropout to apply to attention output,
        by default 0.0
    layer_norm : bool, optional
        Whether to apply layer normalization, by default False
    batch_norm : bool, optional
        Whether to apply batch normalization, by default True
    residual : bool, optional
        Whether to make a residual connection, by default True
    use_bias : bool, optional
        Whether to use bias, by default False
    attention_layer_name : AttentionLayerName, optional
        Choice of the kind of attention layer, by default "GTN"
    non_siblings_map : Optional[Dict[int, Dict[int, List[int]]]], optional
        For each type of graph, maps the edge id to
        all nodes that are not siblings of its source node
    attention_bias : Optional[th.Tensor], optional
        by default None

    Raises
    ------
    ValueError
        Unknown attention type
    ValueError
        out_dim must be divisible by n_heads
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        n_heads: int,
        dropout: float = 0.0,
        layer_norm: bool = False,
        batch_norm: bool = True,
        residual: bool = True,
        use_bias: bool = False,
        attention_layer_name: AttentionLayerName = "GTN",
        non_siblings_map: Optional[Dict[int, Dict[int, List[int]]]] = None,
        attention_bias: Optional[th.Tensor] = None,
    ) -> None:
        super().__init__()
        self.attention_layer_name = attention_layer_name
        self.in_channels = in_dim
        self.out_channels = out_dim
        self.n_heads = n_heads
        self.dropout = dropout
        self.residual = residual
        self.layer_norm = layer_norm
        self.batch_norm = batch_norm
        self.attention_bias = attention_bias
        self.non_siblings_map = non_siblings_map
        attention_layer = ATTENTION_TYPES.get(self.attention_layer_name)
        if not attention_layer:
            raise ValueError(f"Unknown attention type: {self.name}")

        additional_args: Dict[str, Any] = {}
        if attention_layer == QFMultiHeadAttentionLayer:
            additional_args = {"attention_bias": self.attention_bias}
        elif attention_layer == RAALMultiHeadAttentionLayer:
            additional_args = {"non_siblings_map": self.non_siblings_map}
        if out_dim % n_heads != 0:
            raise ValueError(
                f"out_dim ({out_dim}) must be divisible by n_heads ({n_heads})"
            )
        self.attention = attention_layer(
            in_dim,
            out_dim // n_heads,
            n_heads,
            use_bias,
            **additional_args,
        )

        self.O = nn.Linear(out_dim, out_dim)

        if self.layer_norm:
            self.layer_norm1 = nn.LayerNorm(out_dim)

        if self.batch_norm:
            self.batch_norm1 = nn.BatchNorm1d(out_dim)

        # FFN
        self.FFN_layer1 = nn.Linear(out_dim, out_dim * 2)
        self.FFN_layer2 = nn.Linear(out_dim * 2, out_dim)

        if self.layer_norm:
            self.layer_norm2 = nn.LayerNorm(out_dim)

        if self.batch_norm:
            self.batch_norm2 = nn.BatchNorm1d(out_dim)

    def forward(self, g: dgl.DGLGraph, h: th.Tensor) -> th.Tensor:
        h_in1 = h  # for first residual connection
        # multi-head attention out
        attn_out = self.attention.forward(g, h)
        h = attn_out.view(-1, self.out_channels)
        h = F.dropout(h, self.dropout, training=self.training)
        h = self.O(h)
        if self.residual:
            h = h_in1 + h  # residual connection
        if self.layer_norm:
            h = self.layer_norm1(h)
        if self.batch_norm:
            h = self.batch_norm1(h)
        h_in2 = h  # for second residual connection
        # FFN
        h = self.FFN_layer1(h)
        h = F.relu(h)
        h = F.dropout(h, self.dropout, training=self.training)
        h = self.FFN_layer2(h)
        if self.residual:
            h = h_in2 + h  # residual connection
        if self.layer_norm:
            h = self.layer_norm2(h)
        if self.batch_norm:
            h = self.batch_norm2(h)
        return h

    def __repr__(self) -> str:
        return "{}(in_channels={}, out_channels={}, heads={}, residual={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.n_heads,
            self.residual,
        )



from dataclasses import dataclass

import dgl
import torch as th
import torch.nn as nn

from .base_graph_embedder import BaseGraphEmbedder


class GraphAverager(BaseGraphEmbedder):
    """Averager Embedder network.
    Computes an embedding for each operation using a linear layer,
    then averages the embeddings of all operations in the graph.
    """

    @dataclass
    class Params(BaseGraphEmbedder.Params):
        pass

    def __init__(self, net_params: Params) -> None:
        super().__init__(net_params)

        self.emb = nn.Sequential(
            nn.Linear(self.input_size, self.embedding_size), nn.ReLU()
        )

    def _embed(self, g: dgl.DGLGraph, h: th.Tensor) -> th.Tensor:
        h = self.emb(h)
        g.ndata["h"] = h
        return dgl.mean_nodes(g, "h")

    def forward(self, g: dgl.DGLGraph) -> th.Tensor:  # type: ignore[override]
        h = self.concatenate_op_features(g)
        return self.normalize_embedding(self._embed(g, h))



from typing import List, Optional

import pytest
import torch as th

from ....regressors.layers.mlp_readout import MLPReadout
from ....utils.utils import set_deterministic_torch


@pytest.mark.parametrize("agg_dims", [None, [10, 15], [12]])
def test_initialize_with_aggregation_layers(agg_dims: Optional[List[int]]) -> None:
    model = MLPReadout(10, 20, 5, dropout=0, n_layers=3, agg_dims=agg_dims)

    total_layers = 3 + 1  # inner layers + output layer
    if agg_dims:
        total_layers += len(agg_dims)

    assert len(model.FC_layers) == total_layers
    assert model.BN_layers is not None
    assert len(model.BN_layers) == total_layers - 1


def test_init_with_dropout() -> None:
    model = MLPReadout(10, 20, 5, dropout=0.5, n_layers=3, agg_dims=[10, 15])
    assert model.BN_layers is None


@pytest.mark.parametrize(
    "agg_dims, dropout, expected_output",
    [
        (
            None,
            0,
            th.tensor([[-9.637698e-02, -5.446782e-02], [5.620091e-02, -3.493200e-02]]),
        ),
        (
            [10, 15],
            0,
            th.tensor([[-5.963351e-01, 4.445746e-02], [1.587002e-01, 2.566562e-01]]),
        ),
        (
            [12],
            0,
            th.tensor([[-1.248996e-01, -2.419595e-02], [3.839206e-01, -1.896000e-01]]),
        ),
    ],
)
def test_mlp_readout_forward(
    agg_dims: Optional[List[int]], dropout: float, expected_output: th.Tensor
) -> None:
    set_deterministic_torch(0)
    model = MLPReadout(10, 20, 2, dropout=dropout, n_layers=3, agg_dims=agg_dims)
    input_tensor = th.rand((2, 10))
    output = model.forward(input_tensor)
    assert output.shape == (2, 2)
    assert th.allclose(output, expected_output, rtol=1e-5)



from typing import Dict, List, Optional, Tuple

import pandas as pd

from ...utils.logging import logger
from ..containers import QueryStructureContainer
from ..utils.query_plan import (
    QueryPlanOperationFeatures,
    QueryPlanStructure,
    add_positional_encoding,
    extract_query_plan_features,
)
from ..utils.utils import DatasetType
from .base_extractors import TrainedExtractor


class QueryStructureExtractor(TrainedExtractor[QueryStructureContainer]):
    """
    Extracts the features of the operations in the logical plan,
    and the tree structure of the logical plan.
    Keep track of the different query plans seen so far, and their template id.

    Parameters
    ----------
    with_positional_encoding: bool
        Whether to add positional encoding to the query plan gaph.
    """

    def __init__(self, positional_encoding_size: Optional[int] = None) -> None:
        self.template_plans: Dict[int, QueryPlanStructure] = {}
        self.feature_types: Dict[
            str, type
        ] = QueryPlanOperationFeatures.get_feature_names_and_types()
        self.id_template_dict: Dict[str, int] = {}
        self.operation_types: Dict[str, int] = {}
        self.positional_encoding_size = positional_encoding_size

    def _extract_operation_types(
        self, structure: QueryPlanStructure, split: DatasetType
    ) -> List[int]:
        """Find ids of operation types in the query plan.
        Add new operation types if train."""
        operation_gids: List[int] = []
        for name in structure.node_id2name.values():
            op_type = name.split()[0]
            if op_type not in self.operation_types and split == "train":
                self.operation_types[op_type] = len(self.operation_types)
            operation_gids.append(self.operation_types.get(op_type, -1))
        return operation_gids

    def _extract_structure_template(
        self, structure: QueryPlanStructure, split: DatasetType
    ) -> int:
        """Find template id of the query plan, or create a new one (if train)"""
        tid = None
        for template_id, template_structure in self.template_plans.items():
            if structure.graph_match(template_structure):
                tid = template_id
                break

        if tid is None:
            tid = len(self.template_plans) + 1
            if self.positional_encoding_size:
                structure.graph = add_positional_encoding(
                    structure.graph, self.positional_encoding_size
                )
            self.template_plans[tid] = structure
            if split != "train":
                logger.warning(
                    f"Unknown template plan in test or validation set, set as {tid}"
                )
        return tid

    def _extract_structure_and_features(
        self, idx: str, query_plan: str, split: DatasetType
    ) -> Dict:
        """Extract the features of the operations in the logical plan,
        and the tree structure of the logical plan.

        Parameters
        ----------
        idx: str
            The id of the query plan.
        query_plan : str
            A query plan string.

        Returns
        -------
        Dict
            - template_id: id of the template of the query plan
            - operation_id: list of operation ids in the query plan
            - one key per feature for features of the operations
            in the query plan

        """
        structure, op_features, meta_features = extract_query_plan_features(query_plan)
        operation_gids = self._extract_operation_types(structure, split)
        self.id_template_dict[idx] = self._extract_structure_template(structure, split)
        return {
            "operation_id": op_features.operation_ids,
            "operation_gid": operation_gids,
            **op_features.features_dict,
            **meta_features,
        }

    def _derive_meta_dataframe(
        self, df_op_features: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Create meta dataframe from the meta_ columns
        of the df_op_features dataframe. These features are aggregate
        of operations and thusgive one line per query plan."""
        df_meta_features = df_op_features[
            ["plan_id"]
            + [col for col in df_op_features.columns if col.startswith("meta_")]
        ]
        df_meta_features = df_meta_features.rename(
            columns=lambda x: x.replace("meta_", "")
        )
        df_meta_features.set_index("plan_id", inplace=True)
        filtered_df_op_features = df_op_features[
            [col for col in df_op_features.columns if not col.startswith("meta_")]
        ]
        return filtered_df_op_features, df_meta_features

    def _extract_op_features_exploded(
        self, df_op_features: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """Explode the df_op_features dataframe to have one row per operation
        in the query plans, and one column per feature of the operations."
        """
        df_op_features_exploded = df_op_features.explode(
            ["operation_id", "operation_gid"] + list(self.feature_types.keys()),
            ignore_index=True,
        )
        df_op_features_exploded = df_op_features_exploded.set_index(
            ["plan_id", "operation_id"]
        )
        df_op_features_exploded[
            list(self.feature_types.keys())
        ] = df_op_features_exploded[list(self.feature_types.keys())].astype("float32")
        df_operation_types = df_op_features_exploded["operation_gid"].astype("int32")
        del df_op_features_exploded["operation_gid"]
        return df_op_features_exploded, df_operation_types

    def extract_features(
        self, df: pd.DataFrame, split: DatasetType
    ) -> QueryStructureContainer:
        """Extract the features of the operations in the logical plan,
        and the tree structure of the logical plan for each query plan
        in the dataframe.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with a column "plan" containing the query plans.

        Returns
        -------
        pd.DataFrame
            Dataframe with one row per operation in the query plans,
            and one column per feature of the operations.
        """
        df_op_features: pd.DataFrame = df.apply(
            lambda row: self._extract_structure_and_features(row.id, row.plan, split),
            axis=1,
        ).apply(pd.Series)
        df_op_features["plan_id"] = df["id"]
        df_op_features, df_meta_features = self._derive_meta_dataframe(df_op_features)
        (
            df_op_features_exploded,
            df_operation_types,
        ) = self._extract_op_features_exploded(df_op_features)
        return QueryStructureContainer(
            graph_features=df_op_features_exploded,
            template_plans=self.template_plans,
            key_to_template=self.id_template_dict,
            graph_meta_features=df_meta_features,
            operation_types=df_operation_types,
        )



from dataclasses import dataclass

import numpy as np
import pandas as pd

from .base_container import BaseContainer


@dataclass
class TabularContainer(BaseContainer):
    """Container for tabular data, stored in DataFrame format."""

    data: pd.DataFrame

    def get(self, key: str) -> np.ndarray:
        return self.data.loc[key].values  # type: ignore



import dgl
import numpy as np
import pandas as pd
import pytest
import torch as th

from ...containers import QueryStructureContainer, TabularContainer
from ...iterators import QueryPlanIterator
from ...utils.query_plan import QueryPlanStructure


@pytest.fixture
def sample_iterator() -> QueryPlanIterator:
    """Builds a sample QueryPlanIterator, with 3 graphs from 2 templates."""
    keys = ["a", "b", "c"]
    arrays = [["a", "a", "b", "b", "b", "c", "c"], [0, 1, 0, 1, 2, 0, 1]]
    multi_index = pd.MultiIndex.from_arrays(arrays, names=("plan_id", "operation_id"))

    graph_features = pd.DataFrame(
        data=np.vstack([np.linspace(i, i + 1, 2) for i in range(7)]),
        index=multi_index,
        columns=["rows_count", "size"],
    )
    operation_types = pd.DataFrame(
        data=np.vstack([i for i in range(7)]),
        index=multi_index,
        columns=["operation_gid"],
    )["operation_gid"]
    graph_meta_features = pd.DataFrame(
        data=np.vstack([np.linspace(i * 10, (i + 1) * 10, 2) for i in range(3)]),
        index=keys,
        columns=["rows_count", "size"],
    )
    embeddings_features = pd.DataFrame(
        data=np.vstack([np.linspace(i, i + 1, 10) for i in range(7)]),
        index=multi_index,
        columns=[f"emb_{i}" for i in range(10)],
    )
    key_to_template = {"a": 1, "b": 2, "c": 1}

    template_plans = {
        1: QueryPlanStructure(["node_1", "node_2"], [0], [1]),
        2: QueryPlanStructure(["node_1", "node_2", "node_3"], [0, 1], [1, 2]),
    }
    structure_container = QueryStructureContainer(
        graph_features,
        graph_meta_features,
        template_plans,
        key_to_template,
        operation_types=operation_types,
    )

    df_features = pd.DataFrame.from_dict({"id": ["a", "b", "c"], "feature": [1, 2, 1]})
    df_features.set_index("id", inplace=True)
    features_container = TabularContainer(df_features)
    df_objectives = pd.DataFrame.from_dict(
        {"id": ["a", "b", "c"], "objective": [0, 1, 0]}
    )
    df_objectives.set_index("id", inplace=True)
    objectives_container = TabularContainer(df_objectives)

    return QueryPlanIterator(
        keys,
        features_container,
        objectives_container,
        structure_container,
        op_enc=TabularContainer(embeddings_features),
    )


class TestQueryGraphIterator:
    def test_len(self, sample_iterator: QueryPlanIterator) -> None:
        assert len(sample_iterator) == len(sample_iterator.keys)
        assert len(sample_iterator) == len(
            sample_iterator.query_structure_container.key_to_template
        )

    def test_get_item(self, sample_iterator: QueryPlanIterator) -> None:
        first_sample, objectives = sample_iterator[0]

        expected_graph, expected_meta = sample_iterator._get_graph_and_meta("a")
        for key in first_sample.embedding_input.ndata:
            assert th.equal(
                first_sample.embedding_input.ndata[key],  # type: ignore
                expected_graph.ndata[key],  # type: ignore
            )
        assert th.equal(
            first_sample.features,
            th.tensor(np.concatenate([expected_meta, [1]])),
        )
        assert np.equal(objectives, [0])

    def test_get_graph(self, sample_iterator: QueryPlanIterator) -> None:
        graph, meta = sample_iterator._get_graph_and_meta("a")
        assert meta is not None
        assert th.equal(meta, th.tensor([0, 10], dtype=th.float32))
        assert isinstance(graph, dgl.DGLGraph)
        assert graph.number_of_nodes() == 2
        assert th.equal(
            graph.ndata["cbo"],  # type: ignore
            th.tensor(
                np.array([np.linspace(i, i + 1, 2) for i in range(2)]), dtype=th.float32
            ),
        )
        assert th.equal(
            graph.ndata["op_enc"],  # type: ignore
            th.tensor(
                np.array([np.linspace(i, i + 1, 10) for i in range(2)]),
                dtype=th.float32,
            ),
        )

    def test_iterator_shape(self, sample_iterator: QueryPlanIterator) -> None:
        shape = sample_iterator.shape
        assert shape.embedding_input_shape == {"cbo": 2, "op_enc": 10, "type": 7}
        assert shape.feature_names == ["rows_count", "size", "feature"]
        assert shape.output_names == ["objective"]

    def test_set_tensor_type(self, sample_iterator: QueryPlanIterator) -> None:
        sample_iterator.set_tensors_dtype(th.float64)
        assert sample_iterator.tensors_dtype == th.float64
        graph, meta = sample_iterator._get_graph_and_meta("a")
        assert graph.ndata["cbo"].dtype == th.float64  # type: ignore



import dgl
import pytest
import torch as th


def generate_dgl_graph(num_nodes: int, num_edges: int, features: dict) -> dgl.DGLGraph:
    u = th.tensor(range(num_edges))
    v = th.tensor(range(1, num_edges + 1))
    g = dgl.graph((u, v))
    for feat_name, props in features.items():
        if props["type"] == "float":
            g.ndata[feat_name] = th.randn(num_nodes, props["size"])
        elif props["type"] == "int":
            g.ndata[feat_name] = th.randint(0, 3, (num_nodes, props["size"]))
        if props["size"] == 1:
            g.ndata[feat_name] = g.ndata[feat_name].squeeze()  # type: ignore
    return g


@pytest.fixture
def generate_dgl_graph_fixture(request: pytest.FixtureRequest) -> dgl.DGLGraph:
    """
    Generate a DGL graph.

    Parameters
    - num_nodes: number of nodes
    - num_edges: number of edges
    - features: a dictionary with node/edge feature names as keys
                and feature sizes as values
    """
    num_nodes = request.param.get("num_nodes", 2)
    num_edges = request.param.get("num_edges", 2)
    features = request.param.get("features", {})
    g = generate_dgl_graph(num_nodes, num_edges, features)

    return g



import random
import string
from typing import Tuple
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
import torch as th

from ...extractors import TabularFeatureExtractor
from ...handler.data_handler import DataHandler
from ...handler.data_processor import DataProcessor
from ...iterators import TabularIterator


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


@pytest.fixture
def df_fixture() -> Tuple[pd.DataFrame, DataHandler.Params]:
    n = 1000
    ids = list(range(1, n + 1))
    tids = [1 for _ in range(n - 10)] + [2 for _ in range(10)]
    length = [5 for i in range(n)]

    df = pd.DataFrame.from_dict({"id": ids, "tid": tids, "feature": length})

    data_processor = DataProcessor(
        iterator_cls=TabularIterator,
        feature_extractors={"tabular_feature": TabularFeatureExtractor(["feature"])},
    )
    params = DataHandler.Params(
        index_column="id",
        data_processor=data_processor,
        stratify_on=None,
        test_frac=0.1,
        val_frac=0.3,
        random_state=1,
    )
    return df, params


class TestDataHandler:
    def test_split_applies_stratification(
        self, df_fixture: Tuple[pd.DataFrame, DataHandler.Params]
    ) -> None:
        df, params = df_fixture
        params.stratify_on = "tid"
        dh = DataHandler(df, params)
        dh.split_data()
        for split, keys in dh.index_splits.items():
            df_split = df.loc[keys]
            assert len(df_split[df_split["tid"] == 1]) == 99 * len(
                df_split[df_split["tid"] == 2]
            )

    def test_split_no_stratification(
        self, df_fixture: Tuple[pd.DataFrame, DataHandler.Params]
    ) -> None:
        """Check that the split is done correctly
        when no stratification is applied.
        The proportions are correct and there is
        no intersection between the splits."""
        df, params = df_fixture
        params.stratify_on = None
        params.random_state = 1
        dh = DataHandler(df, params)
        dh.split_data()
        assert len(dh.index_splits["test"]) / len(dh.full_df) == dh.test_frac
        assert len(dh.index_splits["val"]) / len(dh.full_df) == dh.val_frac
        assert len(dh.index_splits["train"]) / len(dh.full_df) == 1 - (
            dh.val_frac + dh.test_frac
        )
        assert not set(dh.index_splits["test"]) & set(dh.index_splits["val"])
        assert not set(dh.index_splits["test"]) & set(dh.index_splits["train"])
        assert not set(dh.index_splits["val"]) & set(dh.index_splits["train"])

        df_split = df.loc[dh.index_splits["test"]]
        # results are deterministic because of random_state
        # another random_state would give different results
        assert len(df_split[df_split["tid"] == 1]) == 98
        assert len(df_split[df_split["tid"] == 2]) == 2

    def test_get_iterators_calls_split_data(
        self, df_fixture: Tuple[pd.DataFrame, DataHandler.Params]
    ) -> None:
        df, params = df_fixture
        dh = DataHandler(df, params)
        with patch.object(dh, "split_data") as mock_split_data:
            dh.get_iterators()
            mock_split_data.assert_called_once()

    def test_get_iterators(
        self, df_fixture: Tuple[pd.DataFrame, DataHandler.Params]
    ) -> None:
        np.random.seed(1)
        df, params = df_fixture
        dh = DataHandler(df, params)
        iterators = dh.split_data().get_iterators()
        assert len(iterators) == 3
        assert all(
            isinstance(it, params.data_processor.iterator_cls)
            for it in iterators.values()
        )
        for split, it in iterators.items():
            assert len(it) == len(dh.index_splits[split])
            assert th.eq(it[0], th.tensor(5, dtype=th.float32))



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import torch as th
import torch.optim as optim

from ...data.containers.tabular_container import TabularContainer
from ...data.handler.data_processor import DataProcessor
from ...data.iterators.base_iterator import UdaoIterator
from ...utils.interfaces import UdaoInput, UdaoItemShape
from ...utils.logging import logger
from .. import concepts as co
from ..concepts.utils import derive_processed_input, derive_unprocessed_input
from ..utils.exceptions import NoSolutionError, UncompliantSolutionError
from ..utils.moo_utils import get_default_device
from .so_solver import SOSolver


class MOGD(SOSolver):
    """MOGD solver for single-objective optimization.

    Performs gradient descent on input variables by minimizing an
    objective loss and a constraint loss.
    """

    @dataclass
    class Params:
        learning_rate: float
        """learning rate of Adam optimizer applied to input variables"""
        max_iters: int
        """maximum number of iterations for a single local search"""
        patience: int
        """maximum number of iterations without improvement"""
        multistart: int
        """number of random starts for gradient descent"""
        objective_stress: float = 10.0
        """stress term for objective functions"""
        constraint_stress: float = 1e5
        """stress term for constraint functions"""
        strict_rounding: bool = False
        """whether strictly rounding integer variables at each iteration. """
        batch_size: int = 1
        """batch size for gradient descent"""
        device: Optional[th.device] = field(default_factory=get_default_device)
        """device on which to perform torch operations, by default available device."""
        dtype: th.dtype = th.float32
        """type of the tensors"""

    def __init__(self, params: Params) -> None:
        super().__init__()
        self.lr = params.learning_rate
        self.max_iter = params.max_iters
        self.patience = params.patience
        self.multistart = params.multistart
        self.objective_stress = params.objective_stress
        self.constraint_stress = params.constraint_stress
        self.strict_rounding = params.strict_rounding
        self.batch_size = params.batch_size
        self.device = params.device
        self.dtype = params.dtype

    def _get_unprocessed_input_values(
        self,
        numeric_variables: Dict[str, co.NumericVariable],
        input_parameters: Optional[Dict[str, Any]] = None,
        seed: Optional[int] = None,
    ) -> Tuple[Dict[str, th.Tensor], Dict[str, Any]]:
        """

        Parameters
        ----------
        numeric_variables : Dict[str, co.NumericVariable]
            Numeric variables for which to get random values
        input_parameters : Optional[Dict[str, Any]], optional
            Non decision parts of the input, by default None
        seed : Optional[int], optional
            Random seed, by default None

        Returns
        -------
        Tuple[Dict[str, th.Tensor], Dict[str, Any]]
            - random values as a tensor for each numeric variable
            - input parameters valuies
        """
        numeric_values: Dict[str, np.ndarray] = {}

        for i, (name, variable) in enumerate(numeric_variables.items()):
            numeric_values[name] = co.variable.get_random_variable_values(
                variable, self.batch_size, seed=seed + i if seed is not None else None
            )
        return derive_unprocessed_input(
            input_variables=numeric_values,
            input_parameters=input_parameters,
            device=self.device,
        )

    def _get_processed_input_values(
        self,
        numeric_variables: Dict[str, co.NumericVariable],
        data_processor: DataProcessor,
        input_parameters: Optional[Dict[str, Any]] = None,
        seed: Optional[int] = None,
    ) -> Tuple[UdaoInput, UdaoItemShape, Callable[[th.Tensor], TabularContainer]]:
        """Get random values for numeric variables

        Parameters
        ----------
        numeric_variables : Dict[str, co.NumericVariable]
            Numeric variables on which to apply gradients
        data_processor : DataProcessor
            Data processor to process input variables
        input_parameters : Optional[Dict[str, Any]], optional
            Non decision parts of the input, by default None

        Returns
        -------
        Tuple[UdaoInput, UdaoInputShape, Callable[[th.Tensor], TabularContainer]]
            - random values for numeric variables
            - shape of the input
            - function to convert a tensor to a TabularContainer
        """
        numeric_values: Dict[str, np.ndarray] = {}

        for i, (name, variable) in enumerate(numeric_variables.items()):
            numeric_values[name] = co.variable.get_random_variable_values(
                variable, self.batch_size, seed=seed + i if seed is not None else None
            )
        input_data, iterator = derive_processed_input(
            data_processor=data_processor,
            input_parameters=input_parameters or {},
            input_variables=numeric_values,
            device=self.device,
        )
        make_tabular_container = cast(
            UdaoIterator, iterator
        ).get_tabular_features_container

        input_data_shape = iterator.shape

        return (
            input_data,
            input_data_shape,
            make_tabular_container,
        )

    def _get_unprocessed_input_bounds(
        self,
        numeric_variables: Dict[str, co.NumericVariable],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """

        Parameters
        ----------
        numeric_variables : Dict[str, co.NumericVariable]
            Variables for which to get bounds

        Returns
        -------
        Tuple[Dict[str, float], Dict[str, float]]
            - lower bounds of numeric variables
            - upper bounds of numeric variables
        """
        lower_numeric_values = {
            name: variable.lower for name, variable in numeric_variables.items()
        }
        upper_numeric_values = {
            name: variable.upper for name, variable in numeric_variables.items()
        }
        return lower_numeric_values, upper_numeric_values

    def _get_processed_input_bounds(
        self,
        numeric_variables: Dict[str, co.NumericVariable],
        data_processor: DataProcessor,
        input_parameters: Optional[Dict[str, Any]] = None,
    ) -> Tuple[UdaoInput, UdaoInput]:
        """Get bounds of numeric variables

        Parameters
        ----------
        numeric_variables : Dict[str, co.NumericVariable]
            Numeric variables on which to apply gradients
        data_processor : DataProcessor
            Data processor to process input variables
        input_parameters : Optional[Dict[str, Any]], optional
            Input parameters, by default None

        Returns
        -------
        Tuple[UdaoInput, UdaoInput]
            Lower and upper bounds of numeric
            variables in the form of a UdaoInput
        """
        lower_numeric_values = {
            name: variable.lower for name, variable in numeric_variables.items()
        }
        upper_numeric_values = {
            name: variable.upper for name, variable in numeric_variables.items()
        }
        lower_input, _ = derive_processed_input(
            data_processor=data_processor,
            input_parameters=input_parameters,
            input_variables=lower_numeric_values,
        )
        upper_input, _ = derive_processed_input(
            data_processor=data_processor,
            input_parameters=input_parameters,
            input_variables=upper_numeric_values,
        )
        if self.device:
            return lower_input.to(self.device), upper_input.to(self.device)
        else:
            return lower_input, upper_input

    def _gradient_descent(
        self,
        problem: co.SOProblem,
        input_data: Union[UdaoInput, Dict],
        optimizer: th.optim.Optimizer,
    ) -> Tuple[int, float, float]:
        """Perform a gradient descent step on input variables

        Parameters
        ----------
        problem : co.SOProblem
            Single-objective optimization problem
        input_data : Union[UdaoInput, Dict]
            Input data - can have different types depending on whether
            the input variables are processed or not.
            - UdaoInput: the naive input
            - Dict: {"input_variables": ..., "input_parameters": ...}

        optimizer : th.optim.Optimizer
            PyTorch optimizer

        Returns
        -------
        Tuple[int, float, float]
            - index of minimum loss
            - minimum loss
            - objective value at minimum loss

        Raises
        ------
        UncompliantSolutionError
            If no solution within bounds is found
        """
        # Compute objective, constraints and corresponding losses

        loss_meta = self._compute_loss(problem, input_data)
        sum_loss = loss_meta["sum_loss"]
        min_loss = loss_meta["min_loss"]
        min_loss_id = loss_meta["min_loss_id"]
        best_obj = loss_meta["best_obj"]
        is_within_constraint = loss_meta["is_within_constraint"]

        optimizer.zero_grad()
        sum_loss.backward()  # type: ignore
        optimizer.step()

        if is_within_constraint and (
            self.within_objective_bounds(best_obj, problem.objective)
        ):
            return min_loss_id, min_loss, best_obj
        else:
            raise UncompliantSolutionError("No solution within bounds found!")

    def _log_success(
        self,
        problem: co.SOProblem,
        iter: int,
        best_obj: float,
        best_iter: int,
        best_feature_input: Any,
    ) -> None:
        logger.debug(
            f"Finished at iteration {iter}, best local {problem.objective.name} "
            f"found {best_obj:.5f}"
            f" \nat iteration {best_iter},"
            f" \nwith vars: {best_feature_input}, for "
            f"objective {problem.objective} and constraints {problem.constraints}"
        )

    def _log_failure(
        self,
        problem: co.SOProblem,
        iter: int,
    ) -> None:
        logger.debug(
            f"Finished at iteration {iter}, no valid {problem.objective.name}"
            f" found for input parameters {problem.input_parameters} with "
            f"objective {problem.objective} and constraints {problem.constraints}"
        )

    def _unprocessed_single_start_opt(
        self,
        problem: co.SOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[float, Dict[str, float], float]:
        """Perform a single start optimization, in the case where
        no data processor is defined.
        The input variables are transformed to a dictionary of tensors and are
        optimized directly, by being passed to the objective function along
        with the input parameters.
        """
        best_iter: Optional[int] = None
        best_loss = np.inf
        best_obj: Optional[float] = None
        best_feature_input: Optional[Dict[str, th.Tensor]] = None

        (
            input_variable_values,
            input_parameter_values,
        ) = self._get_unprocessed_input_values(
            cast(Dict[str, co.NumericVariable], problem.variables),
            input_parameters=problem.input_parameters,
            seed=seed,
        )
        lower_input, upper_input = self._get_unprocessed_input_bounds(
            cast(Dict[str, co.NumericVariable], problem.variables)
        )
        for name in input_variable_values:
            input_variable_values[name].requires_grad_(True)
        optimizer = optim.Adam([t for t in input_variable_values.values()], lr=self.lr)
        i = 0
        while i < self.max_iter:
            with th.no_grad():
                input_variable_values_backup = {
                    k: v.detach().clone() for k, v in input_variable_values.items()
                }
            try:
                min_loss_id, min_loss, local_best_obj = self._gradient_descent(
                    problem,
                    {
                        "input_variables": input_variable_values,
                        "input_parameters": input_parameter_values,
                    },
                    optimizer=optimizer,
                )
            except UncompliantSolutionError:
                pass
            else:
                if min_loss < best_loss:
                    best_loss = min_loss
                    best_obj = local_best_obj
                    best_feature_input = {
                        k: v[min_loss_id].reshape(1, -1)
                        for k, v in input_variable_values_backup.items()
                    }
                    best_iter = i

            with th.no_grad():
                # Update input_variable_values with constrained values
                for k in input_variable_values:
                    input_variable_values[k].data = th.clip(
                        input_variable_values[k].data,
                        lower_input[k],
                        upper_input[k],
                    )

                if self.strict_rounding:
                    # Round all integer variables at each iteration
                    for k in input_variable_values:
                        if isinstance(problem.variables[k], co.IntegerVariable):
                            input_variable_values[k].data = input_variable_values[
                                k
                            ].data.round()

            if best_iter is not None and i > best_iter + self.patience:
                break
            i += 1

        if best_iter is None or best_obj is None or best_feature_input is None:
            self._log_failure(problem, i)
            raise NoSolutionError

        if not self.strict_rounding:
            for k in best_feature_input:
                if isinstance(problem.variables[k], co.IntegerVariable):
                    best_feature_input[k].data = best_feature_input[k].data.round()
            loss_meta = self._compute_loss(
                problem,
                {
                    "input_variables": best_feature_input,
                    "input_parameters": input_parameter_values,
                },
            )
            best_loss = loss_meta["min_loss"]
            best_obj = loss_meta["best_obj"]
            is_within_constraint = loss_meta["is_within_constraint"]
            if (
                best_obj is None
                or not is_within_constraint
                or not self.within_objective_bounds(best_obj, problem.objective)
            ):
                self._log_failure(problem, i)
                raise NoSolutionError

        best_raw_vars = {
            name: best_feature_input[name]
            .cpu()
            .numpy()
            .squeeze()
            .tolist()  # turn np.ndarray to float
            for name in problem.variables
        }
        self._log_success(problem, i, best_obj, best_iter, best_raw_vars)
        return best_obj, best_raw_vars, best_loss

    def _processed_single_start_opt(
        self,
        problem: co.SOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[float, Dict[str, float], float]:
        """Perform a single start optimization, in the case where
        a data processor is defined.

        input variables and parameters are processed by the data processor.
        Gradient descent is performed on the processed input variables.
        Variables are then inverse transformed to get the raw variables.
        """
        if not problem.data_processor:
            raise Exception("Data processor is not defined!")
        best_iter: Optional[int] = None
        best_loss = np.inf
        best_obj: Optional[float] = None
        best_feature_input: Optional[th.Tensor] = None
        # Random numeric variables and their characteristics
        (
            input_data,
            input_data_shape,
            make_tabular_container,
        ) = self._get_processed_input_values(
            cast(Dict[str, co.NumericVariable], problem.variables),
            data_processor=problem.data_processor,
            input_parameters=problem.input_parameters,
            seed=seed,
        )
        # Bounds of numeric variables
        lower_input, upper_input = self._get_processed_input_bounds(
            cast(Dict[str, co.NumericVariable], problem.variables),
            data_processor=problem.data_processor,
            input_parameters=problem.input_parameters,
        )
        # Indices of numeric variables on which to apply gradients
        mask = th.tensor(
            [i in problem.variables for i in input_data_shape.feature_names],
            device=self.device,
        )
        grad_indices = th.nonzero(mask, as_tuple=False).squeeze()
        input_vars_subvector = input_data.features[:, grad_indices].clone().detach()
        input_vars_subvector.requires_grad_(True)

        optimizer = optim.Adam([input_vars_subvector], lr=self.lr)
        i = 0
        while i < self.max_iter:
            input_data.features = input_data.features.clone().detach()
            input_data.features[:, grad_indices] = input_vars_subvector
            try:
                min_loss_id, min_loss, local_best_obj = self._gradient_descent(
                    problem,
                    input_data,
                    optimizer=optimizer,
                )
            except UncompliantSolutionError:
                pass
            else:
                if min_loss < best_loss:
                    best_loss = min_loss
                    best_obj = local_best_obj
                    best_feature_input = (
                        input_data.features.detach()[min_loss_id].clone().reshape(1, -1)
                    )
                    best_iter = i

            with th.no_grad():
                # Update input_vars_subvector with constrained values
                input_vars_subvector.data = th.clip(
                    input_vars_subvector.data,
                    # Use .data to avoid gradient tracking during update
                    lower_input.features[0, grad_indices],
                    upper_input.features[0, grad_indices],
                )

                if self.strict_rounding:
                    # Round all integer variables at each iteration
                    input_data.features[:, grad_indices] = input_vars_subvector.data
                    feature_container = make_tabular_container(
                        input_data.features.detach()
                    )
                    best_raw_df = problem.data_processor.inverse_transform(
                        feature_container, "tabular_features"
                    )
                    numeric_values: Dict[str, np.ndarray] = {
                        name: best_raw_df[[name]].values.round()[:, 0]
                        if isinstance(variable, co.IntegerVariable)
                        else best_raw_df[[name]].values[:, 0]
                        for name, variable in problem.variables.items()
                    }
                    input_data_raw, _ = derive_processed_input(
                        data_processor=problem.data_processor,
                        input_parameters=problem.input_parameters or {},
                        input_variables=numeric_values,
                        device=self.device,
                    )
                    input_vars_subvector.data = input_data_raw.features[:, grad_indices]

            if best_iter is not None and i > best_iter + self.patience:
                break
            i += 1

        if best_iter is None or best_obj is None or best_feature_input is None:
            self._log_failure(problem, i)
            raise NoSolutionError

        with th.no_grad():
            best_feature_input = cast(th.Tensor, best_feature_input)
            feature_container = make_tabular_container(best_feature_input)
            best_raw_df = problem.data_processor.inverse_transform(
                feature_container, "tabular_features"
            )
            if not self.strict_rounding:
                best_raw_vars: Dict[str, Any] = {
                    name: best_raw_df[[name]].values.round()[:, 0]
                    if isinstance(variable, co.IntegerVariable)
                    else best_raw_df[[name]].values[:, 0]
                    for name, variable in problem.variables.items()
                }
                input_data_best_raw, _ = derive_processed_input(
                    data_processor=problem.data_processor,
                    input_parameters=problem.input_parameters or {},
                    input_variables=best_raw_vars,
                    device=self.device,
                )
                loss_meta = self._compute_loss(problem, input_data_best_raw)
                best_loss = loss_meta["min_loss"]
                best_obj = loss_meta["best_obj"]
                is_within_constraint = loss_meta["is_within_constraint"]
                if (
                    best_obj is None
                    or not is_within_constraint
                    or not self.within_objective_bounds(best_obj, problem.objective)
                ):
                    self._log_failure(problem, i)
                    raise NoSolutionError
            else:
                best_raw_vars = {
                    name: best_raw_df[[name]]
                    .values.squeeze()
                    .tolist()  # turn np.ndarray to float
                    for name in problem.variables
                }
            self._log_success(problem, i, best_obj, best_iter, best_raw_vars)
            return best_obj, best_raw_vars, best_loss

    def _single_start_opt(
        self,
        problem: co.SOProblem,
        seed: Optional[int] = None,
    ) -> Tuple[float, Dict[str, float], float]:
        """Perform a single start optimization.
        Categorical variables are fixed to the values in input_parameters.
        (a grid search of categorical variables is performed in solve)
        This is where gradient descent is performed.

        Parameters
        ----------
        numeric_variables : Dict[str, co.NumericVariable]
            Numeric variables on which to apply gradients
        objective : co.Objective
            Objective to be optimized
        constraints : Sequence[co.Constraint]
            Constraints to be satisfied
        input_parameters : Optional[Dict[str, Any]], optional
            Non decision parts of the input, by default None
        seed: int, by default None
            random seed

        Returns
        -------
        Tuple[float, Dict[str, float], flat]
            - objective value
            - variables
            - best loss value

        Raises
        ------
        NoSolutionError
            No valid solution is found
        """

        if not problem.data_processor:
            return self._unprocessed_single_start_opt(problem, seed=seed)
        else:
            return self._processed_single_start_opt(problem, seed=seed)

    def solve(
        self, problem: co.SOProblem, seed: Optional[int] = None
    ) -> Tuple[float, Dict[str, float]]:
        if seed is not None:
            th.manual_seed(seed)
        if self.device:
            for constraint in problem.constraints:
                constraint.to(self.device)
            problem.objective.to(self.device)

        categorical_variables = [
            name
            for name, variable in problem.variables.items()
            if isinstance(variable, co.EnumVariable)
        ]
        numeric_variables = {
            name: variable
            for name, variable in problem.variables.items()
            if isinstance(variable, co.NumericVariable)
        }

        meshed_categorical_vars = self.get_meshed_categorical_vars(problem.variables)

        if meshed_categorical_vars is None:
            meshed_categorical_vars = np.array([0])

        best_loss_list: List[float] = []
        obj_list: List[float] = []
        vars_list: List[Dict] = []
        for i in range(self.multistart):
            for categorical_cell in meshed_categorical_vars:
                categorical_values = {
                    name: categorical_cell[ind]
                    for ind, name in enumerate(categorical_variables)
                }  # from {id: value} to {name: value}
                fixed_values = {
                    **categorical_values,
                    **(problem.input_parameters or {}),
                }
                try:
                    (
                        obj_pred,
                        best_raw_vars,
                        best_loss,
                    ) = self._single_start_opt(
                        co.SOProblem(
                            variables=numeric_variables,  # type: ignore
                            input_parameters=fixed_values,
                            objective=problem.objective,
                            constraints=problem.constraints or [],
                            data_processor=problem.data_processor,
                        ),
                        seed=seed + i if seed is not None else None,
                    )
                except NoSolutionError:
                    continue
                else:
                    best_loss_list.append(best_loss)
                    obj_list.append(obj_pred)
                    vars_list.append(best_raw_vars)
        if not obj_list:
            raise NoSolutionError("No valid solutions and variables found!")

        idx = np.argmin(best_loss_list)
        vars_cand = vars_list[idx]
        if vars_cand is not None:
            obj_cand = obj_list[idx]
            if obj_cand is None:
                raise Exception(f"Unexpected objs_list[{idx}] is None.")
        else:
            raise NoSolutionError("No valid solutions and variables found!")

        return obj_cand, vars_cand

    ##################
    ## _loss        ##
    ##################
    def constraints_loss(
        self, constraint_values: List[th.Tensor], constraints: Sequence[co.Constraint]
    ) -> th.Tensor:
        """
        compute loss of the values of each constraint function fixme: double-check

        Parameters
        ----------
        constraint_values : List[th.Tensor]
            values of each constraint function
        constraints : Sequence[co.Constraint]
            constraint functions

        Returns
        -------
        th.Tensor
            loss of the values of each constraint function

        """

        # vars: a tensor
        # get loss for constraint functions defined in the problem setting
        total_loss = th.zeros_like(
            constraint_values[0], device=self.device, dtype=self.dtype
        )
        for i, (constraint_value, constraint) in enumerate(
            zip(constraint_values, constraints)
        ):
            stress = (
                self.objective_stress
                if isinstance(constraint, co.Objective)
                else self.constraint_stress
            )
            constraint_violation = th.zeros_like(
                constraint_values[0], device=self.device, dtype=self.dtype
            )
            if constraint.upper is not None and constraint.lower is not None:
                if constraint.upper == constraint.lower:
                    constraint_violation = th.abs(constraint_value - constraint.upper)
                else:
                    normed_constraint = (constraint_value - constraint.lower) / (
                        constraint.upper - constraint.lower
                    )
                    constraint_violation = th.where(
                        (normed_constraint < 0) | (normed_constraint > 1),
                        (normed_constraint - 0.5),
                        0,
                    )
            elif constraint.lower is not None:
                constraint_violation = th.relu(constraint.lower - constraint_value)
            elif constraint.upper is not None:
                constraint_violation = th.relu(constraint_value - constraint.upper)
            total_loss += (
                constraint_violation**2 + stress * (constraint_violation > 0).float()
            )

        return total_loss

    def objective_loss(
        self, objective_value: th.Tensor, objective: co.Objective
    ) -> th.Tensor:
        """Compute the objective loss for a given objective value:
        - if no bounds are specified, use the squared objective value
        - if both bounds are specified, use the squared normalized
        objective value if it is within the bounds, otherwise
        add a stress term to a squared distance to middle of the bounds

        Parameters
        ----------
        objective_value : th.Tensor
            Tensor of objective values
        objective : co.Objective
            Objective function

        Returns
        -------
        th.Tensor
            Tensor of objective losses

        Raises
        ------
        NotImplementedError
            If only one bound is specified for the objective

        """

        if objective.upper is None and objective.lower is None:
            loss = (
                th.sign(objective_value) * (objective_value**2) * objective.direction
            )
        elif objective.upper is not None and objective.lower is not None:
            norm_cst_obj_pred = (objective_value - objective.lower) / (
                objective.upper - objective.lower
            )  # scaled
            loss = th.where(
                (norm_cst_obj_pred < 0) | (norm_cst_obj_pred > 1),
                (norm_cst_obj_pred - 0.5) ** 2 + self.objective_stress,
                norm_cst_obj_pred * objective.direction,
            )
        else:
            raise NotImplementedError("Objective with only one bound is not supported")
        return loss

    def _obj_forward(
        self,
        optimization_element: co.Constraint,
        input_data: Union[UdaoInput, Dict],
    ) -> th.Tensor:
        if isinstance(input_data, UdaoInput):
            return optimization_element.function(input_data)  # type: ignore
        else:
            # Dict when unprocessed inputs
            return optimization_element.function(**input_data)

    def _compute_loss(
        self, problem: co.SOProblem, input_data: Union[UdaoInput, Dict]
    ) -> Dict[str, Any]:
        obj_output = self._obj_forward(problem.objective, input_data)
        objective_loss = self.objective_loss(obj_output, problem.objective)
        constraint_loss = th.zeros_like(objective_loss, device=self.device)

        if problem.constraints:
            const_outputs = [
                self._obj_forward(constraint, input_data)
                for constraint in problem.constraints
            ]
            constraint_loss = self.constraints_loss(const_outputs, problem.constraints)

        loss = objective_loss + constraint_loss
        min_loss_id = int(th.argmin(loss).cpu().item())

        return {
            "sum_loss": th.sum(loss),
            "min_loss": th.min(loss).cpu().item(),
            "min_loss_id": min_loss_id,
            "best_obj": obj_output[min_loss_id].cpu().item(),
            "is_within_constraint": bool((constraint_loss[min_loss_id] == 0).item()),
        }

    ##################
    ## _get (vars)  ##
    ##################

    def get_meshed_categorical_vars(
        self, variables: Dict[str, co.Variable]
    ) -> Optional[np.ndarray]:
        """
        Get combinations of all categorical (binary, enum) variables

        Parameters
        ----------
        variables : Dict[str, co.Variable]
            Variables to be optimized

        Returns
        -------
        Optional[np.ndarray]
            Combinations of all categorical variables
            of shape (n_samples, n_vars)
        """
        cv_value_list = [
            variable.values
            for variable in variables.values()
            if isinstance(variable, co.EnumVariable)
        ]
        if not cv_value_list:
            return None
        meshed_cv_value_list = [x_.reshape(-1, 1) for x_ in np.meshgrid(*cv_value_list)]
        meshed_cv_value = np.concatenate(meshed_cv_value_list, axis=1)
        return meshed_cv_value

    ##################
    ## _check       ##
    ##################

    @staticmethod
    def within_objective_bounds(obj_value: float, objective: co.Objective) -> bool:
        """
        check whether violating the objective value var_ranges
        :param pred_dict: dict, keys are objective names,
        values are objective values
        :param obj_bounds: dict, keys are objective names,
        values are lower and upper var_ranges of each objective value
        :return: True or False
        """
        within_bounds = True
        if objective.upper is not None:
            within_bounds = obj_value <= objective.upper
        if objective.lower is not None:
            within_bounds = within_bounds and obj_value >= objective.lower
        return within_bounds



from .base_progressive_frontier import BaseProgressiveFrontier
from .parallel_progressive_frontier import ParallelProgressiveFrontier
from .sequential_progressive_frontier import SequentialProgressiveFrontier

__all__ = [
    "BaseProgressiveFrontier",
    "ParallelProgressiveFrontier",
    "SequentialProgressiveFrontier",
]



from dataclasses import dataclass, field
from typing import Dict, Mapping, Optional

import numpy as np
import torch as th

from ..concepts import EnumVariable, FloatVariable, IntegerVariable, Variable
from ..utils.moo_utils import get_default_device
from .sampler_solver import SamplerSolver


class RandomSamplerSolver(SamplerSolver):
    """Solving a SOO problem by random sampling over variables"""

    @dataclass
    class Params:
        n_samples_per_param: int
        "the number of samples per variable"
        device: Optional[th.device] = field(default_factory=get_default_device)
        """device on which to perform torch operations, by default available device."""

    def __init__(self, params: Params) -> None:
        super().__init__(params.device)
        self.n_samples_per_param = params.n_samples_per_param

    def _process_variable(
        self, var: Variable, seed: Optional[int] = None
    ) -> np.ndarray:
        if seed is not None:
            np.random.seed(seed)
        """Generate samples of a variable"""
        if isinstance(var, FloatVariable):
            return np.random.uniform(var.lower, var.upper, self.n_samples_per_param)
        elif isinstance(var, IntegerVariable):
            return np.random.randint(
                var.lower, var.upper + 1, size=self.n_samples_per_param
            )
        elif isinstance(var, EnumVariable):
            inds = np.random.randint(0, len(var.values), size=self.n_samples_per_param)
            return np.array(var.values)[inds]
        else:
            raise NotImplementedError(
                f"ERROR: variable type {type(var)} is not supported!"
            )

    def _get_input(
        self, variables: Mapping[str, Variable], seed: Optional[int] = None
    ) -> Dict[str, np.ndarray]:
        """
        generate samples of variables

        Parameters:
        -----------
        variables: List[Variable],
            lower and upper var_ranges of variables(non-ENUM),
            and values of ENUM variables
        Returns:
        --------
        Dict[str, np.ndarray]
            Dict with array of values for each variable
        """
        result_dict = {}

        for name, var in variables.items():
            result_dict[name] = self._process_variable(var, seed=seed)

        return result_dict



import re
from collections import defaultdict
from typing import Callable, Dict, List, Tuple

import pandas as pd


def remove_unknown(s: str) -> str:
    """Remove unknown symbol from a query plan
    (in the form of (unknown))
    """
    pattern = r"\(unknown\)"
    # Remove unknown operations
    s = re.sub(pattern, "", s)
    return s


def remove_statistics(s: str) -> str:
    """Remove statistical information from a query plan
    (in the form of Statistics(...)
    """
    pattern = r"\bStatistics\([^)]+\)"
    # Remove statistical information
    s = re.sub(pattern, "", s)
    return s


def remove_hashes(s: str) -> str:
    """Remove hashes from a query plan, e.g. #1234L"""
    # Replace hashes with a placeholder or remove them
    return re.sub(r"#[0-9]+[L]*", "", s)


def brief_clean(s: str) -> str:
    """Remove special characters from a string and convert to lower case"""
    return re.sub(r"[^0-9A-Za-z\'_.]+", " ", s).lower()


def replace_symbols(s: str) -> str:
    """Replace symbols with tokens"""
    return (
        s.replace(" >= ", " GE ")
        .replace(" <= ", " LE ")
        .replace(" == ", " EQ")
        .replace(" = ", " EQ ")
        .replace(" > ", " GT ")
        .replace(" < ", " LT ")
        .replace(" != ", " NEQ ")
        .replace(" + ", " rADD ")
        .replace(" - ", " rMINUS ")
        .replace(" / ", " rDIV ")
        .replace(" * ", " rMUL ")
    )


def remove_duplicate_spaces(s: str) -> str:
    return " ".join(s.split())


def prepare_operation(operation: str) -> str:
    """Prepare an operation for embedding by keeping only
    relevant semantic information"""
    processings: List[Callable[[str], str]] = [
        remove_unknown,
        remove_statistics,
        remove_hashes,
        replace_symbols,
        brief_clean,
        remove_duplicate_spaces,
    ]
    for processing in processings:
        operation = processing(operation)
    return operation


def build_unique_operations(df: pd.DataFrame) -> Tuple[Dict[int, List[int]], List[str]]:
    """Build a dictionary of unique operations and their IDs"""
    unique_ops: Dict[str, int] = defaultdict(lambda: len(unique_ops))
    plan_to_ops: Dict[int, List[int]] = defaultdict(list)
    for row in df.itertuples():
        plan_to_ops[row.id].append(unique_ops[row.operation])

    operations_list = list(unique_ops.keys())
    return plan_to_ops, operations_list


def extract_operations(
    plan_df: pd.DataFrame, operation_processing: Callable[[str], str] = lambda x: x
) -> Tuple[Dict[int, List[int]], List[str]]:
    """Extract unique operations from a DataFrame of
    query plans and links them to query plans.
    Operations are transformed using prepare_operation
    to remove statistical information and hashes.

    Parameters
    ----------
    plan_df : pd.DataFrame
        DataFrame containing the query plans and their ids.

    operation_processing : Callable[[str], str]
        Function to process the operations, by default no processing will be applied
        and the raw operations will be used.

    Returns
    -------
    Tuple[Dict[int, List[int]], List[str]]
        plan_to_ops: Dict[int, List[int]]
            Links a query plan ID to a list of operation IDs in the operations list
        operations_list: List[str]
            List of unique operations in the dataset
    """
    df = plan_df[["id", "plan"]].copy()

    df["plan"] = df["plan"].apply(
        lambda plan: [operation_processing(op) for op in plan.splitlines()]  # type: ignore
    )
    df = df.explode("plan", ignore_index=True)
    df.rename(columns={"plan": "operation"}, inplace=True)
    return build_unique_operations(df)



from .base_iterator import BaseIterator, UdaoIterator
from .query_plan_iterator import QueryPlanIterator
from .tabular_iterator import TabularIterator

__all__ = [
    "BaseIterator",
    "UdaoIterator",
    "TabularIterator",
    "QueryPlanIterator",
]



from typing import Any, Callable, Optional, Union

import torch as th

from .utils import UdaoFunction


class Constraint:
    """An optimization element is either an objective or a constraint.

    The choice of the type depends on whether a DataProcessor is specified
    for the problem:
    - if no DataProcessor is provided: UdaoFunction, it is a callable
    that takes input_variables and input_parameters
    - else, th.nn.Module or other Callable returning a tensor.

    Parameters
    ----------
    function : Union[UdaoFunction, th.nn.Module, Callable[..., th.Tensor]]
        Objective function, either a UdaoFunction
        or a th.nn.Module if a DataProcessor is provided
    lower : Optional[float], optional
        lower bound of the element, by default None
    upper : Optional[float], optional
        upper bound of the element, by default None
    """

    def __init__(
        self,
        function: Union[UdaoFunction, th.nn.Module, Callable[..., th.Tensor]],
        lower: Optional[float] = None,
        upper: Optional[float] = None,
    ) -> None:
        if isinstance(function, th.nn.Module):
            function.eval()
            for p in function.parameters():
                p.requires_grad = False
        self.function = function
        self.lower = lower
        self.upper = upper

    def __call__(self, *args: Any, **kwargs: Any) -> th.Tensor:
        return self.function(*args, **kwargs)

    def to(self, device: Optional[th.device]) -> "Constraint":
        if isinstance(self.function, th.nn.Module) and device is not None:
            self.function.to(device)
        return self

    def __repr__(self) -> str:
        return f"Constraint(lower={self.lower}, upper={self.upper})"



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional

import pandas as pd
import torch as th

from ...utils.logging import logger
from ..containers import BaseContainer
from ..iterators import BaseIterator
from ..utils.utils import DatasetType, train_test_val_split_on_column
from .data_processor import DataProcessor


class DataHandler:
    """
    DataHandler class to handle data loading, splitting, feature extraction and
    dataset iterator creation.

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe containing the data.
    params : DataHandler.Params
        DataHandler.Params object containing the parameters of the DataHandler.
    """

    @dataclass
    class Params:
        index_column: str
        """Column that should be used as index (unique identifier)"""

        data_processor: DataProcessor
        """DataProcessor object to extract features from the data and create the
            iterator."""

        stratify_on: Optional[str] = None
        """Column on which to stratify the split, by default None.
        If None, no stratification is performed."""

        val_frac: float = 0.2
        """Column on which to stratify the split
            (keeping proportions for each split)
            If None, no stratification is performed"""

        test_frac: float = 0.1
        """Fraction allotted to the validation set, by default 0.2"""

        dryrun: bool = False
        """Dry run mode for fast computation on a large dataset (sampling of a
            small portion), by default False"""

        random_state: Optional[int] = None
        """Random state for reproducibility, by default None"""

        tensors_dtype: Optional[th.dtype] = None
        """Data type of the tensors, by default None"""

    @classmethod
    def from_csv(
        cls,
        csv_path: str,
        params: Params,
    ) -> "DataHandler":
        """Initialize DataHandler from csv.

        Parameters
        ----------
        csv_path : str
            Path to the data file.
        params : Params
        Returns
        -------
        DataHandler
            Initialized DataHandler object.
        """
        return cls(
            pd.read_csv(csv_path),
            params,
        )

    def __init__(
        self,
        data: pd.DataFrame,
        params: Params,
    ) -> None:
        self.dryrun = params.dryrun
        self.index_column = params.index_column
        self.data_processor = params.data_processor
        self.stratify_on = params.stratify_on
        self.val_frac = params.val_frac
        self.test_frac = params.test_frac
        self.random_state = params.random_state
        self.tensors_dtype = params.tensors_dtype
        self.full_df = data
        if self.dryrun:
            self.full_df = self.full_df.sample(
                frac=0.05, random_state=self.random_state
            )
        self.full_df.set_index(self.index_column, inplace=True, drop=False)
        self.splits: List[DatasetType] = ["train", "val", "test"]
        self.index_splits: Dict[DatasetType, List[str]] = {}
        self.features: Dict[DatasetType, Dict[str, BaseContainer]] = defaultdict(dict)

    def split_data(
        self,
    ) -> "DataHandler":
        """Split the data into train, test and validation sets,
        split indices are stored in self.index_splits.

        Returns
        -------
        DataHandler
            set
        """

        df_splits = train_test_val_split_on_column(
            self.full_df,
            self.stratify_on,
            val_frac=self.val_frac,
            test_frac=self.test_frac,
            random_state=self.random_state,
        )
        self.index_splits = {
            split: df.index.to_list() for split, df in df_splits.items()
        }
        return self

    def _get_split_iterator(self, split: DatasetType) -> BaseIterator:
        keys = self.index_splits[split]
        iterator = self.data_processor.make_iterator(
            keys=self.index_splits[split], data=self.full_df.loc[keys], split=split
        )
        return iterator

    def get_iterators(self) -> Dict[DatasetType, BaseIterator]:
        """Return a dictionary of iterators for the different splits of the data.

        Returns
        -------
        Dict[DatasetType, BaseDatasetIterator]
            Dictionary of iterators for the different splits of the data.
        """
        if not self.index_splits:
            logger.warning("No data split yet. Splitting data now.")
            self.split_data()
        return {split: self._get_split_iterator(split) for split in self.index_splits}



from typing import cast

import numpy as np
import pytest
import torch as th

from ....data.handler.data_processor import DataProcessor
from ....model.utils.utils import set_deterministic_torch
from ...concepts.problem import MOProblem
from ...moo.progressive_frontier import ParallelProgressiveFrontier
from ...soo.mogd import MOGD
from ...utils.moo_utils import Point, Rectangle


@pytest.fixture
def ppf(data_processor: DataProcessor, mogd: MOGD) -> ParallelProgressiveFrontier:
    ppf = ParallelProgressiveFrontier(
        params=ParallelProgressiveFrontier.Params(
            processes=1,
            n_grids=2,
            max_iters=4,
        ),
        solver=mogd,
    )
    return ppf


class TestParallelProgressiveFrontier:
    def test_create_grid_cells(self, ppf: ParallelProgressiveFrontier) -> None:
        utopia = Point(np.array([0, 2, 0]))
        nadir = Point(np.array([4, 10, 1]))
        grid_rectangles = ppf._create_grid_cells(utopia, nadir, 2, 3)

        assert len(grid_rectangles) == 8
        expected = [
            Rectangle(
                utopia=Point(objs=np.array([0.0, 2.0, 0.0])),
                nadir=Point(objs=np.array([2.0, 6.0, 0.5])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([0.0, 2.0, 0.5])),
                nadir=Point(objs=np.array([2.0, 6.0, 1.0])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([0.0, 6.0, 0.0])),
                nadir=Point(objs=np.array([2.0, 10.0, 0.5])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([0.0, 6.0, 0.5])),
                nadir=Point(objs=np.array([2.0, 10.0, 1.0])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([2.0, 2.0, 0.0])),
                nadir=Point(objs=np.array([4.0, 6.0, 0.5])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([2.0, 2.0, 0.5])),
                nadir=Point(objs=np.array([4.0, 6.0, 1.0])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([2.0, 6.0, 0.0])),
                nadir=Point(objs=np.array([4.0, 10.0, 0.5])),
            ),
            Rectangle(
                utopia=Point(objs=np.array([2.0, 6.0, 0.5])),
                nadir=Point(objs=np.array([4.0, 10.0, 1.0])),
            ),
        ]
        for i, rect in enumerate(expected):
            assert rect == grid_rectangles[i]

    def test_solve_with_two_objectives(
        self, ppf: ParallelProgressiveFrontier, two_obj_problem: MOProblem
    ) -> None:
        set_deterministic_torch(0)
        objectives, variables = ppf.solve(
            problem=two_obj_problem,
            seed=0,
        )
        assert objectives is not None
        cast(MOGD, ppf.solver).patience = 100
        np.testing.assert_array_equal(objectives, np.array([[0, 0]]))
        assert variables is not None
        assert variables[0] == {"v1": 0.0, "v2": 1.0}

    def test_solve_with_two_objectives_gpu(
        self, ppf: ParallelProgressiveFrontier, two_obj_problem: MOProblem
    ) -> None:
        if not th.cuda.is_available():
            pytest.skip("Skip GPU test")
        cast(MOGD, ppf.solver).device = th.device("cuda")
        set_deterministic_torch(0)
        objectives, variables = ppf.solve(
            problem=two_obj_problem,
            seed=0,
        )
        assert objectives is not None
        cast(MOGD, ppf.solver).patience = 100
        np.testing.assert_array_equal(objectives, np.array([[0, 0]]))
        assert variables is not None
        assert variables[0] == {"v1": 0.0, "v2": 1.0}

    def test_solve_with_three_objectives(
        self, ppf: ParallelProgressiveFrontier, three_obj_problem: MOProblem
    ) -> None:
        set_deterministic_torch()
        obj_values, var_values = ppf.solve(problem=three_obj_problem, seed=0)
        assert obj_values is not None
        np.testing.assert_array_almost_equal(obj_values, np.array([[-1.0, -1.0, -2.0]]))
        assert var_values[0] == {"v1": 1.0, "v2": 7.0}



from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, Tuple

import dgl
import torch as th

from ...data.containers.tabular_container import TabularContainer
from ...utils.interfaces import UdaoEmbedInput, UdaoEmbedItemShape
from ..containers import QueryStructureContainer
from .base_iterator import UdaoIterator


@dataclass
class QueryPlanInput(UdaoEmbedInput[dgl.DGLGraph]):
    """The embedding input is a dgl.DGLGraph"""

    def to(self, device: th.device) -> "QueryPlanInput":
        return QueryPlanInput(self.features.to(device), self.embedding_input.to(device))


class QueryPlanIterator(UdaoIterator[QueryPlanInput, UdaoEmbedItemShape]):
    """
    Iterator that returns a dgl.DGLGraph for each key, with associated node features.
    The features are stored in the graph.ndata dictionary.
    The features are expected to be float tensors, and to be of the same length
    as the number of nodes in the graph.

    Parameters
    ----------
    keys : Sequence[str]
        Keys of the dataset, used for accessing all features
    tabular_features : TabularContainer
        Container for the tabular features associated with the plan
    objectives : TabularContainer
        Container for the objectives associated with the plan
    query_structure : QueryStructureContainer
        Wrapper around the graph structure and the features for each query plan
    kwargs: BaseContainer
        Variable number of other features to add to the graph, e.g. embeddings
    """

    def __init__(
        self,
        keys: Sequence[str],
        tabular_features: TabularContainer,
        objectives: TabularContainer,
        query_structure: QueryStructureContainer,
        **kwargs: TabularContainer,
    ):
        super().__init__(
            keys=keys,
            tabular_features=tabular_features,
            objectives=objectives,
        )
        self.query_structure_container = query_structure
        self.base_graph_features = ["cbo"]
        self.other_graph_features = kwargs

    def __len__(self) -> int:
        return len(self.keys)

    def _get_graph_and_meta(self, key: str) -> Tuple[dgl.DGLGraph, Optional[th.Tensor]]:
        """Returns the graph corresponding to the key,
        associated with features as th.tensor,
        and the meta information.
        """
        query = self.query_structure_container.get(key)
        graph = query.template_graph
        graph.ndata["cbo"] = th.tensor(query.graph_features, dtype=self.tensors_dtype)
        graph.ndata["op_gid"] = th.tensor(query.operation_types, dtype=th.int32)
        for feature, container in self.other_graph_features.items():
            graph.ndata[feature] = th.tensor(
                container.get(key), dtype=self.tensors_dtype
            )
        meta_input = (
            None
            if query.meta_features is None
            else th.tensor(query.meta_features, dtype=self.tensors_dtype)
        )
        return graph, meta_input

    def _getitem(self, idx: int) -> Tuple[QueryPlanInput, th.Tensor]:
        key = self.keys[idx]
        features = th.tensor(self.tabular_features.get(key), dtype=self.tensors_dtype)
        objectives = th.tensor(self.objectives.get(key), dtype=self.tensors_dtype)
        graph, meta_input = self._get_graph_and_meta(key)
        if meta_input is not None:
            features = th.cat([meta_input, features])
        input_data = QueryPlanInput(features, graph)
        return input_data, objectives

    @property
    def shape(self) -> UdaoEmbedItemShape[Dict[str, int]]:
        """Returns the dimensions of the iterator inputs and outputs."""

        sample_input, sample_output = self._get_sample()
        embedding_input_shape = {}
        graph_feature_names = self.base_graph_features + list(
            self.other_graph_features.keys()
        )
        for feature_name in graph_feature_names:
            embedding_input_shape[feature_name] = sample_input.embedding_input.ndata[
                feature_name
            ].shape[  # type: ignore
                1
            ]
        embedding_input_shape["type"] = len(
            self.query_structure_container.operation_types.unique()
        )
        meta_features = self.query_structure_container.graph_meta_features
        feature_names = self.tabular_features.data.columns.tolist()
        if meta_features is not None:
            feature_names[:0] = meta_features.columns.tolist()
        output_names = list(self.objectives.data.columns)
        return UdaoEmbedItemShape(
            embedding_input_shape=embedding_input_shape,
            feature_names=feature_names,
            output_names=output_names,
        )

    @staticmethod
    def collate(
        items: List[Tuple[QueryPlanInput, th.Tensor]],
    ) -> Tuple[QueryPlanInput, th.Tensor]:
        """Collate a list of FeatureItem into a single graph."""
        graphs = [item[0].embedding_input for item in items]
        features = th.vstack([item[0].features for item in items])
        objectives = th.vstack([item[1] for item in items])
        return QueryPlanInput(features, dgl.batch(graphs)), objectives

    @staticmethod
    def make_graph_augmentation(
        augmentation: Callable[[dgl.DGLGraph], dgl.DGLGraph],
    ) -> Callable:
        def graph_augmentation(
            item: Tuple[QueryPlanInput, th.Tensor]
        ) -> Tuple[QueryPlanInput, th.Tensor]:
            query_plan, output = item
            query_plan.embedding_input = augmentation(query_plan.embedding_input)
            return query_plan, output

        return graph_augmentation



from typing import Dict, Iterable, Optional

import numpy as np
import pytest
import torch as th

from ... import concepts as co
from ...soo.random_sampler_solver import RandomSamplerSolver


class TestRandomSampler:
    @pytest.mark.parametrize(
        "test_data, expected",
        [
            ({"variable": co.BoolVariable(), "n_samples": 3}, [0, 1, 1]),
            (
                {"variable": co.IntegerVariable(1, 7), "n_samples": 2},
                [5, 6],
            ),
            (
                {"variable": co.FloatVariable(2, 4), "n_samples": 5},
                [3.098, 3.430, 3.206, 3.090, 2.847],
            ),
            (
                {
                    "variable": co.EnumVariable([0, 4, 7, 10]),
                    "n_samples": 2,
                    "range": [0, 4, 7, 10],
                },
                [0, 10],
            ),
        ],
    )
    def test_random_sampler_single_variable(
        self, test_data: Dict, expected: Iterable
    ) -> None:
        solver = RandomSamplerSolver(
            RandomSamplerSolver.Params(n_samples_per_param=test_data["n_samples"])
        )
        output = solver._get_input(variables={"v1": test_data["variable"]}, seed=0)
        np.testing.assert_allclose(
            [[o] for o in output["v1"]], [[e] for e in expected], rtol=1e-2
        )

    def test_random_sampler_multiple_variables(self) -> None:
        solver = RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=3))

        output = solver._get_input(
            variables={"v1": co.BoolVariable(), "v2": co.IntegerVariable(1, 7)}, seed=0
        )
        expected_array = np.array(
            [
                [0, 5],
                [1, 6],
                [1, 1],
            ]
        )
        np.testing.assert_equal(
            output, {"v1": expected_array[:, 0], "v2": expected_array[:, 1]}
        )

    def test_solve(self) -> None:
        solver = RandomSamplerSolver(RandomSamplerSolver.Params(n_samples_per_param=50))

        def obj1_func(
            input_variables: Dict[str, th.Tensor],
            input_parameters: Optional[Dict[str, th.Tensor]] = None,
        ) -> th.Tensor:
            return input_variables["v1"] + input_variables["v2"]

        objective = co.Objective("obj1", minimize=False, function=obj1_func)
        variables: Dict[str, co.Variable] = {
            "v1": co.BoolVariable(),
            "v2": co.IntegerVariable(1, 7),
        }
        problem = co.SOProblem(objective=objective, variables=variables, constraints=[])
        soo_obj, soo_vars = solver.solve(problem, seed=0)
        assert soo_obj == 8
        assert soo_vars == {"v1": 1, "v2": 7}



from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence

import torch as th

from ...data.handler.data_processor import DataProcessor
from ..concepts.utils import (
    InputVariables,
    derive_processed_input,
    derive_unprocessed_input,
)
from .constraint import Constraint
from .objective import Objective
from .variable import Variable


class BaseProblem:
    """Base class for optimization problems."""

    def __init__(
        self,
        variables: Dict[str, Variable],
        constraints: Sequence[Constraint],
        data_processor: Optional[DataProcessor] = None,
        input_parameters: Optional[Dict[str, Any]] = None,
    ) -> None:
        """_summary_

        Parameters
        ----------
        variables : Dict[str, Variable]
            _description_
        constraints : Sequence[Constraint]
            _description_
        data_processor : Optional[DataProcessor], optional
            _description_, by default None
        input_parameters : Optional[Dict[str, Any]], optional
            _description_, by default None
        """
        self.variables = variables
        self.constraints = constraints
        self.data_processor = data_processor
        self.input_parameters = input_parameters

    def apply_function(
        self,
        optimization_element: Constraint,
        input_variables: InputVariables,
        device: Optional[th.device] = None,
    ) -> th.Tensor:
        if self.data_processor is not None:
            input_data, _ = derive_processed_input(
                self.data_processor,
                input_variables=input_variables,
                input_parameters=self.input_parameters,
                device=device,
            )
            th_value = optimization_element.to(device)(input_data)
        else:
            input_vars, input_params = derive_unprocessed_input(
                input_variables,
                self.input_parameters,
                device=device,
            )
            th_value = optimization_element(
                input_parameters=input_params,
                input_variables=input_vars,
            )

        return th_value

    def derive_SO_problem(self, objective: Objective) -> "SOProblem":
        return SOProblem(
            objective=objective,
            variables=self.variables,
            constraints=self.constraints,
            data_processor=self.data_processor,
            input_parameters=self.input_parameters,
        )


@dataclass
class MOProblem(BaseProblem):
    """Multi-objective optimization problem."""

    def __init__(
        self,
        objectives: Sequence[Objective],
        variables: Dict[str, Variable],
        constraints: Sequence[Constraint],
        data_processor: Optional[DataProcessor] = None,
        input_parameters: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.objectives = objectives
        super().__init__(
            variables,
            constraints,
            data_processor=data_processor,
            input_parameters=input_parameters,
        )

    def __repr__(self) -> str:
        return (
            f"MOProblem(objectives={self.objectives}, "
            f"variables={self.variables}, "
            f"constraints={self.constraints}, "
            f"input_parameters={self.input_parameters})"
        )


@dataclass
class SOProblem(BaseProblem):
    """Single-objective optimization problem."""

    def __init__(
        self,
        objective: Objective,
        variables: Dict[str, Variable],
        constraints: Sequence[Constraint],
        data_processor: Optional[DataProcessor] = None,
        input_parameters: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.objective = objective
        super().__init__(
            variables,
            constraints,
            data_processor=data_processor,
            input_parameters=input_parameters,
        )

    def __repr__(self) -> str:
        return (
            f"SOProblem(objective={self.objective}, "
            f"variables={self.variables}, "
            f"constraints={self.constraints}, "
            f"input_parameters={self.input_parameters})"
        )



from .base_embedder import BaseEmbedder
from .base_graph_embedder import BaseGraphEmbedder
from .graph_averager import GraphAverager
from .graph_transformer import GraphTransformer

__all__ = [
    "BaseEmbedder",
    "BaseGraphEmbedder",
    "GraphAverager",
    "GraphTransformer",
]



from typing import Any, Dict, List, Sequence

import torch as th

from ..containers import TabularContainer
from .base_iterator import BaseIterator


class TabularIterator(BaseIterator[th.Tensor, Dict[str, Any]]):
    """Iterator on tabular data.

    Parameters
    ----------
    keys : Sequence[str]
        Keys of the dataset, used for accessing all features
    table : TabularContainer
        Container for the tabular data
    """

    def __init__(
        self,
        keys: Sequence[str],
        tabular_feature: TabularContainer,
    ):
        super().__init__(keys)
        self.tabular_feature = tabular_feature

    def __len__(self) -> int:
        return len(self.keys)

    def _getitem(self, idx: int) -> th.Tensor:
        key = self.keys[idx]
        return th.tensor(self.tabular_feature.get(key), dtype=self.tensors_dtype)

    @property
    def shape(self) -> Any:
        sample_input = self._get_sample()
        return {"input_shape": sample_input.shape}

    @staticmethod
    def collate(items: List[th.Tensor]) -> th.Tensor:
        return th.stack(items, dim=0)



from typing import Any, List, Protocol, TypeVar

import pandas as pd

from ..containers.base_container import BaseContainer
from ..utils.utils import DatasetType
from .base_preprocessor import TrainedPreprocessor


# Define a protocol for objects that have fit and transform methods
class FitTransformProtocol(Protocol):
    def fit(self, X: Any, y: Any = None) -> "FitTransformProtocol":
        ...

    def transform(self, X: Any) -> Any:
        ...

    def inverse_transform(self, X: Any) -> Any:
        ...


T = TypeVar("T", bound=BaseContainer)


class NormalizePreprocessor(TrainedPreprocessor[T]):
    """Normalize the data using a normalizer that
    implements the fit and transform methods, e.g. MinMaxScaler.

    Parameters
    ----------
    normalizer : FitTransformProtocol
        A normalizer that implements the fit and transform methods
        (e.g. sklearn.MinMaxScaler)
    df_key : str
        The key of the dataframe in the container.
    """

    def __init__(
        self, normalizer: FitTransformProtocol, data_key: str = "data"
    ) -> None:
        self.normalizer = normalizer
        self.df_key = data_key
        self.numeric_columns: List[str] = []

    def preprocess(self, container: T, split: DatasetType) -> T:
        """Normalize the data in the container.

        Parameters
        ----------
        container : T
            Child of BaseContainer
        split : DatasetType
            Train or other (val, test).
        Returns
        -------
        T
            Child of BaseContainer with the normalized data.
        """
        container = container.copy()
        df: pd.DataFrame = container.__getattribute__(self.df_key)
        numeric_df = df.select_dtypes(include=["number"])
        self.numeric_columns = numeric_df.columns.tolist()
        if split == "train":
            self.normalizer.fit(numeric_df)

        transformed_data = self.normalizer.transform(numeric_df)
        transformed_df = pd.DataFrame(
            transformed_data, index=numeric_df.index, columns=self.numeric_columns
        )

        df.update(transformed_df)

        container.__setattr__(self.df_key, df)
        return container

    def inverse_transform(self, container: T) -> T:
        """Reverse the normalization process on the container's data.

        Parameters
        ----------
        container : T
            Child of BaseContainer with the normalized data.

        Returns
        -------
        T
            Child of BaseContainer with the data in original scale.
        """
        container = container.copy()
        df: pd.DataFrame = container.__getattribute__(self.df_key)
        numeric_df = df[self.numeric_columns]

        original_data = self.normalizer.inverse_transform(numeric_df.values)
        original_numeric_df = pd.DataFrame(
            original_data, index=numeric_df.index, columns=self.numeric_columns
        )

        df.update(original_numeric_df)

        container.__setattr__(self.df_key, df)
        return container



import copy
from abc import ABC, abstractmethod
from typing import TypeVar

T = TypeVar("T", bound="BaseContainer")


class BaseContainer(ABC):
    """Base class for containers.
    Containers are used to store and retrieve data
    from a dataset, based on a key."""

    @abstractmethod
    def get(self, key: str):  # type: ignore
        pass

    def copy(self: T) -> T:
        return copy.deepcopy(self)



from typing import Dict, List, Optional, Union

import pandas as pd

from ...utils.interfaces import VarTypes
from ..containers import TabularContainer
from .base_extractors import StaticExtractor


class TabularFeatureExtractor(StaticExtractor[TabularContainer]):
    """
    Extract columns from a DataFrame as a TabularContainer.

    Parameters
    ----------
    columns : Union[List[str], Dict[str, Optional[VarTypes]]], optional
        Either:
        - a list of column names to extract from the DataFrame
        - a dictionary that maps column names to variable types
        if the variable type is None, the column is extracted
        without casting
        - None, in which case all columns are extracted

    """

    def __init__(
        self, columns: Optional[Union[List[str], Dict[str, Optional[VarTypes]]]] = None
    ) -> None:
        self.columns = columns

    def extract_features(self, df: pd.DataFrame) -> TabularContainer:
        """extract and cast features from the DataFrame"""
        if isinstance(self.columns, list):
            extracted_df = df[self.columns]
        elif isinstance(self.columns, dict):
            extracted_df = pd.DataFrame()
            for col, var_type in self.columns.items():
                if isinstance(var_type, VarTypes):
                    extracted_df[col] = df[col].astype(var_type.value)
                elif var_type is None:
                    extracted_df[col] = df[col]
                else:
                    raise Exception(f"Unknown variable type: {var_type}")
        else:
            extracted_df = df

        return TabularContainer(extracted_df)



"""Provides a logger object for the udao library."""
import logging
import sys


def _get_logger(name: str = "udao", level: int = logging.DEBUG) -> logging.Logger:
    """Generates a logger object for the UDAO library.

    Parameters
    ----------
    name : str, optional
        logger name, by default "udao".
    level : int, optional
        logging level (DEBUG, INFO...), by default logging.DEBUG

    Returns
    -------
    logging.Logger
        logger object to call for logging
    """
    _logger = logging.getLogger(name)
    _logger.setLevel(level)
    if not _logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(level)
        log_format = (
            "%(asctime)s - [%(levelname)s] - %(name)s - "
            "(%(filename)s:%(lineno)d) - "
            "%(message)s"
        )
        formatter = logging.Formatter(log_format)
        handler.setFormatter(formatter)

        _logger.addHandler(handler)
    return _logger


logger = _get_logger()



from dataclasses import dataclass
from typing import List, Optional

import torch as th

from .base_regressor import BaseRegressor
from .layers.mlp_readout import MLPReadout


class MLP(BaseRegressor):
    """MLP to compute the final regression from
    the embedding and the tabular input features.
    Parameters
    ----------
    net_params : MLP.Params
        For the parameters, see the MLP.Params dataclass.
    """

    @dataclass
    class Params(BaseRegressor.Params):
        n_layers: int
        """Number of layers in the MLP"""
        hidden_dim: int
        """Size of the hidden layers outputs."""
        dropout: float
        """Probability of dropout."""
        agg_dims: Optional[List[int]] = None
        """Dimensions of the aggregation layers in the MLP."""

    def __init__(self, net_params: Params) -> None:
        """_summary_"""
        super().__init__(net_params)

        self.MLP_layers = MLPReadout(
            input_dim=self.input_dim,
            hidden_dim=net_params.hidden_dim,
            output_dim=net_params.output_dim,
            n_layers=net_params.n_layers,
            dropout=net_params.dropout,
            agg_dims=net_params.agg_dims,
        )

    def forward(self, embedding: th.Tensor, inst_feat: th.Tensor) -> th.Tensor:
        hgi = th.cat([embedding, inst_feat], dim=1)
        out = self.MLP_layers.forward(hgi)
        return th.exp(out)



import json
from pathlib import Path
from typing import Dict, NamedTuple

import pytest

from ..utils.query_plan import (
    QueryPlanOperationFeatures,
    QueryPlanStructure,
    compute_meta_features,
    format_size,
)

QueryPlanElements = NamedTuple(
    "QueryPlanElements",
    [
        ("query_plan", str),
        ("structure", QueryPlanStructure),
        ("features", QueryPlanOperationFeatures),
        ("meta_features", Dict[str, float]),
    ],
)


def get_query_plan_sample(
    json_path: str,
) -> QueryPlanElements:
    base_dir = Path(__file__).parent
    with open(base_dir / json_path, "r") as f:
        plan_features = json.load(f)
    incoming_ids = plan_features["incoming_ids"]
    outgoing_ids = plan_features["outgoing_ids"]
    names = plan_features["names"]
    row_counts = [float(v) for v in plan_features["row_counts"]]
    sizes = [format_size(s) for s in plan_features["sizes"]]
    structure = QueryPlanStructure(names, incoming_ids, outgoing_ids)
    operation_features = QueryPlanOperationFeatures(row_counts, sizes)
    meta_features = compute_meta_features(structure, operation_features)
    return QueryPlanElements(
        plan_features["query_plan"], structure, operation_features, meta_features
    )


@pytest.fixture(scope="session")
def sample_plan_1() -> QueryPlanElements:
    base_dir = Path(__file__).parent
    return get_query_plan_sample(str(base_dir / "assets/sample_plan_1.json"))


@pytest.fixture(scope="session")
def sample_plan_2() -> QueryPlanElements:
    base_dir = Path(__file__).parent
    return get_query_plan_sample(str(base_dir / "assets/sample_plan_2.json"))



from abc import ABC
from dataclasses import dataclass
from typing import Any, Optional, Tuple

import numpy as np
import torch as th

from ....utils.interfaces import VarTypes
from ....utils.logging import logger
from ...concepts import Objective
from ...concepts.problem import MOProblem, SOProblem
from ...soo.so_solver import SOSolver
from ...utils.exceptions import NoSolutionError
from ...utils.moo_utils import Point
from ..mo_solver import MOSolver


class BaseProgressiveFrontier(MOSolver, ABC):
    """
    Base class for Progressive Frontier.
    Includes the common methods for Progressive Frontier.
    """

    @dataclass
    class Params:
        """Parameters for Progressive Frontier"""

        pass

    def __init__(
        self,
        solver: SOSolver,
        params: Params,
    ) -> None:
        super().__init__()
        self.solver = solver
        self.opt_obj_ind = 0

    def get_anchor_point(
        self,
        problem: MOProblem,
        obj_ind: int,
        seed: Optional[int] = None,
    ) -> Point:
        """
        Find the anchor point for the given objective,
        by unbounded single objective optimization

        Parameters
        ----------
        problem : MOProblem
            MOO problem in which the objective is to be optimized
        obj_ind : int
            index of the objective to be optimized
        Returns
        -------
        Point
            anchor point for the given objective
        """
        try:
            _, soo_vars = self.solver.solve(
                problem.derive_SO_problem(
                    objective=problem.objectives[obj_ind],
                ),
                seed=seed,
            )
        except NoSolutionError:
            raise NoSolutionError("Cannot find anchor points.")
        else:
            objs = self._compute_objectives(problem, soo_vars)

        # If the current objective type is Integer,
        # further find the optimal value for other objectives with float type
        if problem.objectives[obj_ind].type == VarTypes.INT:
            utopia_init = np.array(
                [0 if i != obj_ind else objs[obj_ind] for i in problem.objectives]
            )
            utopia_tmp, nadir_tmp = Point(objs=utopia_init), Point(objs=objs)
            # select the first objective with float type
            float_obj_ind = [
                i
                for i, objective in enumerate(problem.objectives)
                if objective.type == VarTypes.FLOAT
            ][0]
            obj_bounds_dict_so = self._form_obj_bounds_dict(
                problem, utopia_tmp, nadir_tmp
            )
            so_problem = self._so_problem_from_bounds_dict(
                problem, obj_bounds_dict_so, problem.objectives[float_obj_ind]
            )
            try:
                _, soo_vars_update = self.solver.solve(so_problem, seed=seed)
            except NoSolutionError:
                raise NoSolutionError("Cannot find anchor points.")
            else:
                logger.debug(f"soo_vars_update is: {soo_vars_update}")
                objs = self._compute_objectives(problem, soo_vars_update)

                return Point(objs, soo_vars_update)
        else:
            return Point(objs, soo_vars)

    def _form_obj_bounds_dict(
        self, problem: MOProblem, utopia: Point, nadir: Point
    ) -> dict[str, list[float]]:
        """
        form the dict used in the constrained optimization
        e.g. the format:
        obj_bounds_dict = {
            "latency": [0, 10000000],
            "cores": [0, 58]
            }
        Parameters
        ----------
        utopia: Point
            the utopia point
        nadir: Point
            the nadir point
        opt_obj_ind: int
            the index of objective to be optimized

        Returns
        -------
            dict with upper and lower bound for each objective
        """
        bounds = {}
        for i, objective in enumerate(problem.objectives):
            if objective.direction < 0:
                bounds[objective.name] = [
                    nadir.objs[i] * objective.direction,
                    utopia.objs[i] * objective.direction,
                ]
            else:
                bounds[objective.name] = [
                    utopia.objs[i] * objective.direction,
                    nadir.objs[i] * objective.direction,
                ]

        return bounds

    def _so_problem_from_bounds_dict(
        self,
        problem: MOProblem,
        obj_bounds_dict: dict[str, list],
        primary_obj: Objective,
    ) -> SOProblem:
        """

        Parameters
        ----------
        obj_bounds_dict : dict[str, list]
            A lower and upper bound for each objective
        primary_obj : Objective
            The objective to be optimized

        Returns
        -------
        Tuple[Objective, Sequence[Constraint]]
            The objective and constraints for the single-objective optimization
        """
        soo_objective = Objective(
            name=primary_obj.name,
            minimize=primary_obj.minimize,
            function=primary_obj.function,
            lower=obj_bounds_dict[primary_obj.name][0],
            upper=obj_bounds_dict[primary_obj.name][1],
        )
        so_problem = problem.derive_SO_problem(soo_objective)
        soo_constraints = list(so_problem.constraints)

        for obj in problem.objectives:
            obj_name = obj.name
            if obj_name != primary_obj.name:
                soo_constraints.append(
                    Objective(
                        name=obj.name,
                        minimize=obj.minimize,
                        lower=obj_bounds_dict[obj_name][0],
                        upper=obj_bounds_dict[obj_name][1],
                        function=obj.function,
                    )
                )
        so_problem.constraints = soo_constraints

        return so_problem

    @staticmethod
    def get_utopia_and_nadir(points: list[Point]) -> Tuple[Point, Point]:
        """
        get the utopia and nadir points from a list of points
        Parameters
        ----------
        points: list[Point],
            each element is a Point (defined class).

        Returns
        -------
        Tuple[Point, Point]
            utopia and nadir point
        """
        if len(points) == 0:
            raise ValueError("The input list of points is empty.")
        n_objs = points[0].n_objs
        if any([point.n_objs != n_objs for point in points]):
            raise Exception("The number of objectives is not consistent among points.")
        best_objs = [min([point.objs[i] for point in points]) for i in range(n_objs)]
        worst_objs = [max([point.objs[i] for point in points]) for i in range(n_objs)]
        logger.debug(f"best_objs {best_objs}")
        utopia = Point(np.array(best_objs))
        nadir = Point(np.array(worst_objs))

        return utopia, nadir

    def _compute_objectives(
        self,
        problem: MOProblem,
        variable_values: dict[str, Any],
    ) -> np.ndarray:
        """Compute an array of objective for a given point.
        (variable_values is a dict of variable name and single value)

        Parameters
        ----------
        variable_values : dict[str, Any]
            Name: value of variables
        input_parameters : Optional[dict[str, Any]]
            Name: value of other fixed input parameters

        Returns
        -------
        np.ndarray
            _description_
        """
        obj_list = []
        for obj in problem.objectives:
            obj_value = problem.apply_function(
                obj, variable_values, device=th.device("cpu")
            )
            obj_value = (obj_value * obj.direction).squeeze()
            obj_list.append(obj_value.detach())
        return np.array(obj_list)



from .base_extractors import FeatureExtractor, StaticExtractor, TrainedExtractor
from .predicate_embedding_extractor import PredicateEmbeddingExtractor
from .query_structure_extractor import QueryStructureExtractor
from .tabular_extractor import TabularFeatureExtractor

__all__ = [
    "FeatureExtractor",
    "StaticExtractor",
    "TrainedExtractor",
    "PredicateEmbeddingExtractor",
    "QueryStructureExtractor",
    "TabularFeatureExtractor",
]
